"""
awslimitchecker/tests/services/test_vpc.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
from awslimitchecker.tests.services import result_fixtures
from awslimitchecker.services.vpc import _VpcService

from botocore.exceptions import ClientError

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import patch, call, Mock, DEFAULT
else:
    from unittest.mock import patch, call, Mock, DEFAULT


class Test_VpcService(object):

    pb = 'awslimitchecker.services.vpc._VpcService'  # patch base path
    pbm = 'awslimitchecker.services.vpc'  # patch base path - module

    def test_init(self):
        """test __init__()"""
        cls = _VpcService(21, 43, {}, None)
        assert cls.service_name == 'VPC'
        assert cls.conn is None
        assert cls.warning_threshold == 21
        assert cls.critical_threshold == 43

    def test_get_limits(self):
        cls = _VpcService(21, 43, {}, None)
        cls.limits = {}
        res = cls.get_limits()
        assert sorted(res.keys()) == sorted([
            'Entries per route table',
            'Internet gateways',
            'VPCs',
            'Subnets per VPC',
            'NAT Gateways per AZ',
            'Network ACLs per VPC',
            'Rules per network ACL',
            'Route tables per VPC',
            'Virtual private gateways',
            'Network interfaces per Region',
        ])
        for name, limit in res.items():
            assert limit.service == cls
            assert limit.def_warning_threshold == 21
            assert limit.def_critical_threshold == 43

    def test_get_limits_again(self):
        """test that existing limits dict is returned on subsequent calls"""
        mock_limits = Mock()
        cls = _VpcService(21, 43, {}, None)
        cls.limits = mock_limits
        res = cls.get_limits()
        assert res == mock_limits

    def test_find_usage(self):
        mock_conn = Mock()
        sn = {
            'sn-1': 'az1',
            'sn-2': 'az2'
        }

        with patch('%s.connect' % self.pb) as mock_connect:
            with patch.multiple(
                    self.pb,
                    _find_usage_vpcs=DEFAULT,
                    _find_usage_subnets=DEFAULT,
                    _find_usage_ACLs=DEFAULT,
                    _find_usage_route_tables=DEFAULT,
                    _find_usage_gateways=DEFAULT,
                    _find_usage_nat_gateways=DEFAULT,
                    _find_usages_vpn_gateways=DEFAULT,
                    _find_usage_network_interfaces=DEFAULT,
            ) as mocks:
                mocks['_find_usage_subnets'].return_value = sn
                cls = _VpcService(21, 43, {}, None)
                cls.conn = mock_conn
                assert cls._have_usage is False
                cls.find_usage()
        assert mock_connect.mock_calls == [call()]
        assert cls._have_usage is True
        assert mock_conn.mock_calls == []
        for x in [
                '_find_usage_vpcs',
                '_find_usage_subnets',
                '_find_usage_ACLs',
                '_find_usage_route_tables',
                '_find_usage_gateways',
                '_find_usages_vpn_gateways',
                '_find_usage_network_interfaces',
        ]:
            assert mocks[x].mock_calls == [call()]
        assert mocks['_find_usage_nat_gateways'].mock_calls == [call(sn)]

    def test_find_usage_vpcs(self):
        response = result_fixtures.VPC.test_find_usage_vpcs

        mock_conn = Mock()
        mock_conn.describe_vpcs.return_value = response

        cls = _VpcService(21, 43, {}, None)
        cls._current_account_id = '0123456789'
        cls.conn = mock_conn

        cls._find_usage_vpcs()

        assert len(cls.limits['VPCs'].get_current_usage()) == 1
        assert cls.limits['VPCs'].get_current_usage()[0].get_value() == 2
        assert mock_conn.mock_calls == [
            call.describe_vpcs(Filters=[{
                'Name': 'owner-id', 'Values': ['0123456789']
            }])
        ]

    def test_find_usage_subnets(self):
        response = result_fixtures.VPC.test_find_usage_subnets

        mock_conn = Mock()
        mock_conn.describe_subnets.return_value = response
        cls = _VpcService(21, 43, {}, None)
        cls._current_account_id = '0123456789'
        cls.conn = mock_conn

        res = cls._find_usage_subnets()
        assert res == {
            'string': 'string',
            'subnet2': 'az3',
            'subnet3': 'az2'
        }

        usage = sorted(cls.limits['Subnets per VPC'].get_current_usage())
        assert len(usage) == 2
        assert usage[0].get_value() == 1
        assert usage[0].resource_id == 'vpc-2'
        assert usage[1].get_value() == 2
        assert usage[1].resource_id == 'vpc-1'
        assert mock_conn.mock_calls == [
            call.describe_subnets(Filters=[{
                'Name': 'owner-id', 'Values': ['0123456789']
            }])
        ]

    def test_find_usage_acls(self):
        response = result_fixtures.VPC.test_find_usage_acls
        mock_conn = Mock()

        cls = _VpcService(21, 43, {}, None)
        cls._current_account_id = '0123456789'
        cls.conn = mock_conn

        mock_conn.describe_network_acls.return_value = response
        cls._find_usage_ACLs()

        usage = sorted(cls.limits['Network ACLs per VPC'].get_current_usage())
        assert len(usage) == 2
        assert usage[0].get_value() == 1
        assert usage[0].resource_id == 'vpc-2'
        assert usage[1].get_value() == 2
        assert usage[1].resource_id == 'vpc-1'
        entries = sorted(cls.limits['Rules per network '
                                    'ACL'].get_current_usage())
        assert len(entries) == 3
        assert entries[0].resource_id == 'acl-2'
        assert entries[0].get_value() == 1
        assert entries[1].resource_id == 'acl-1'
        assert entries[1].get_value() == 3
        assert entries[2].resource_id == 'acl-3'
        assert entries[2].get_value() == 5
        assert mock_conn.mock_calls == [
            call.describe_network_acls(Filters=[{
                'Name': 'owner-id', 'Values': ['0123456789']
            }])
        ]

    def test_find_usage_route_tables(self):
        response = result_fixtures.VPC.test_find_usage_route_tables

        mock_conn = Mock()
        mock_conn.describe_route_tables.return_value = response

        cls = _VpcService(21, 43, {}, None)
        cls._current_account_id = '0123456789'
        cls.conn = mock_conn

        cls._find_usage_route_tables()

        usage = sorted(cls.limits['Route tables per VPC'].get_current_usage())
        assert len(usage) == 2
        assert usage[0].get_value() == 1
        assert usage[0].resource_id == 'vpc-2'
        assert usage[1].get_value() == 2
        assert usage[1].resource_id == 'vpc-1'
        entries = sorted(cls.limits['Entries per route '
                                    'table'].get_current_usage())
        assert len(entries) == 3
        assert entries[0].resource_id == 'rt-2'
        assert entries[0].get_value() == 1
        assert entries[1].resource_id == 'rt-1'
        assert entries[1].get_value() == 2
        assert entries[2].resource_id == 'rt-3'
        assert entries[2].get_value() == 3
        assert mock_conn.mock_calls == [
            call.describe_route_tables(Filters=[{
                'Name': 'owner-id', 'Values': ['0123456789']
            }])
        ]

    def test_find_usage_internet_gateways(self):
        response = result_fixtures.VPC.test_find_usage_internet_gateways

        mock_conn = Mock()
        mock_conn.describe_internet_gateways.return_value = response

        cls = _VpcService(21, 43, {}, None)
        cls._current_account_id = '0123456789'
        cls.conn = mock_conn

        cls._find_usage_gateways()

        assert len(cls.limits['Internet gateways'].get_current_usage()) == 1
        assert cls.limits['Internet gateways'].get_current_usage()[
            0].get_value() == 2
        assert mock_conn.mock_calls == [
            call.describe_internet_gateways(Filters=[{
                'Name': 'owner-id', 'Values': ['0123456789']
            }])
        ]

    def test_find_usage_nat_gateways(self):
        subnets = result_fixtures.VPC.test_find_usage_nat_gateways_subnets
        response = result_fixtures.VPC.test_find_usage_nat_gateways

        mock_conn = Mock()
        mock_conn.describe_nat_gateways.return_value = response

        with patch('%s.logger' % self.pbm) as mock_logger:
            cls = _VpcService(21, 43, {}, None)
            cls._current_account_id = '0123456789'
            cls.conn = mock_conn
            cls._find_usage_nat_gateways(subnets)

        assert len(cls.limits['NAT Gateways per AZ'].get_current_usage()) == 2
        az2 = cls.limits['NAT Gateways per AZ'].get_current_usage()[0]
        assert az2.get_value() == 3
        assert az2.resource_id == 'az2'
        az3 = cls.limits['NAT Gateways per AZ'].get_current_usage()[1]
        assert az3.get_value() == 1
        assert az3.resource_id == 'az3'
        assert mock_conn.mock_calls == [
            call.describe_nat_gateways(),
        ]
        assert mock_logger.mock_calls == [
            call.error(
                'ERROR: NAT Gateway %s in SubnetId %s, but SubnetId not '
                'found in subnet_to_az; Gateway cannot be counted!',
                'nat-124', 'subnet4'
            ),
            call.debug(
                'Skipping NAT Gateway %s in state: %s', 'nat-125', 'deleted'
            ),
            call.debug(
                'Skipping NAT Gateway %s in state: %s', 'nat-127', 'failed'
            )
        ]

    def test_find_usage_nat_gateways_exception(self):
        subnets = result_fixtures.VPC.test_find_usage_nat_gateways_subnets

        def se_exc(*args, **kwargs):
            raise ClientError({'Error': {}}, 'opname')

        mock_conn = Mock()
        mock_conn.describe_nat_gateways.side_effect = se_exc

        cls = _VpcService(21, 43, {}, None)
        cls._current_account_id = '0123456789'
        cls.conn = mock_conn

        with patch('%s.logger' % self.pbm, autospec=True) as mock_logger:
            cls._find_usage_nat_gateways(subnets)

        assert len(cls.limits['NAT Gateways per AZ'].get_current_usage()) == 0
        assert mock_conn.mock_calls == [
            call.describe_nat_gateways(),
        ]
        assert mock_logger.mock_calls == [
            call.error('Caught exception when trying to list NAT Gateways; '
                       'perhaps NAT service does not exist in this region?',
                       exc_info=1)
        ]

    def test_find_usages_vpn_gateways(self):
        response = result_fixtures.VPC.test_find_usages_vpn_gateways

        mock_conn = Mock()
        mock_conn.describe_vpn_gateways.return_value = response

        cls = _VpcService(21, 43, {}, None)
        cls._current_account_id = '0123456789'
        cls.conn = mock_conn

        cls._find_usages_vpn_gateways()

        assert len(cls.limits['Virtual private gateways']
                   .get_current_usage()) == 1
        assert cls.limits['Virtual private gateways'].get_current_usage()[
            0].get_value() == 2
        assert mock_conn.mock_calls == [
            call.describe_vpn_gateways(Filters=[
                {
                    'Name': 'state',
                    'Values': [
                        'available',
                        'pending'
                    ]
                }
            ]),
        ]

    def test_find_usage_network_interfaces(self):
        response = result_fixtures.VPC.test_find_usage_network_interfaces

        mock_conn = Mock()
        mock_conn.describe_network_interfaces.return_value = response

        cls = _VpcService(21, 43, {}, None)
        cls._current_account_id = '0123456789'
        cls.conn = mock_conn

        cls._find_usage_network_interfaces()

        assert len(cls.limits['Network interfaces per Region']
                   .get_current_usage()) == 1
        assert cls.limits['Network interfaces per Region'].get_current_usage()[
            0].get_value() == 1
        assert mock_conn.mock_calls == [
            call.describe_network_interfaces(Filters=[{
                'Name': 'owner-id', 'Values': ['0123456789']
            }]),
        ]

    def test_required_iam_permissions(self):
        cls = _VpcService(21, 43, {}, None)
        assert cls.required_iam_permissions() == [
            'ec2:DescribeNatGateways',
            'ec2:DescribeNetworkAcls',
            'ec2:DescribeRouteTables',
            'ec2:DescribeSubnets',
            'ec2:DescribeVpcs',
            'ec2:DescribeVpnGateways',
            'ec2:DescribeNetworkInterfaces',
        ]



"""
awslimitchecker/metrics/dummy.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from awslimitchecker.metrics.base import MetricsProvider

logger = logging.getLogger(__name__)


class Dummy(MetricsProvider):
    """Just writes metrics to STDOUT; mainly used for testing."""

    def __init__(self, region_name, **_):
        super(Dummy, self).__init__(region_name)

    def flush(self):
        print('DummyMetrics Provider flush for region=%s' % self._region_name)
        print('Duration: %s' % self._duration)
        lines = []
        for lim in self._limits:
            u = lim.get_current_usage()
            if len(u) == 0:
                max_usage = 0
            else:
                max_usage = max(u).get_value()
            limit = lim.get_limit()
            if limit is None:
                limit = 'unknown'
            lines.append(
                '%s / %s: limit=%s max_usage=%s' % (
                    lim.service.service_name, lim.name, limit, max_usage
                )
            )
        for l in sorted(lines):
            print(l)



"""
awslimitchecker/metrics/datadog.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import os
import logging
import urllib3
import time
import re
import json
from awslimitchecker.metrics.base import MetricsProvider

logger = logging.getLogger(__name__)


class Datadog(MetricsProvider):
    """Send metrics to Datadog."""

    def __init__(
        self, region_name, prefix='awslimitchecker.', api_key=None,
        extra_tags=None
    ):
        """
        Initialize the Datadog metrics provider. This class does not have any
        additional requirements. You must specify at least the ``api_key``
        configuration option.

        :param region_name: the name of the region we're connected to. This
          parameter is automatically passed in by the Runner class.
        :type region_name: str
        :param prefix: Datadog metric prefix
        :type prefix: str
        :param api_key: Datadog API key. May alternatively be specified by the
          ``DATADOG_API_KEY`` environment variable.
        :type api_key: str
        :param extra_tags: CSV list of additional tags to send with metrics.
          All metrics will automatically be tagged with ``region:<region name>``
        :type extra_tags: str
        """
        super(Datadog, self).__init__(region_name)
        self._prefix = prefix
        self._tags = ['region:%s' % region_name]
        if extra_tags is not None:
            self._tags.extend(extra_tags.split(','))
        self._api_key = os.environ.get('DATADOG_API_KEY')
        if api_key is not None:
            self._api_key = api_key
        if self._api_key is None:
            raise RuntimeError(
                'ERROR: Datadog metrics provider requires datadog API key.'
            )
        self._http = urllib3.PoolManager()
        self._validate_auth(self._api_key)

    def _validate_auth(self, api_key):
        url = 'https://api.datadoghq.com/api/v1/validate?api_key=%s'
        logger.debug('Validating Datadog API key: GET %s', url)
        url = url % api_key
        r = self._http.request('GET', url)
        if r.status != 200:
            raise RuntimeError(
                'ERROR: Datadog API key validation failed with HTTP %s: %s' % (
                    r.status, r.data
                )
            )

    def _name_for_metric(self, service, limit):
        """
        Return a metric name that's safe for datadog

        :param service: service name
        :type service: str
        :param limit: limit name
        :type limit: str
        :return: datadog metric name
        :rtype: str
        """
        return ('%s%s.%s' % (
            self._prefix,
            re.sub(r'[^0-9a-zA-Z]+', '_', service),
            re.sub(r'[^0-9a-zA-Z]+', '_', limit)
        )).lower()

    def flush(self):
        ts = int(time.time())
        logger.debug('Flushing metrics to Datadog.')
        series = [{
            'metric': '%sruntime' % self._prefix,
            'points': [[ts, self._duration]],
            'type': 'gauge',
            'tags': self._tags
        }]
        for lim in self._limits:
            u = lim.get_current_usage()
            if len(u) == 0:
                max_usage = 0
            else:
                max_usage = max(u).get_value()
            mname = self._name_for_metric(lim.service.service_name, lim.name)
            series.append({
                'metric': '%s.max_usage' % mname,
                'points': [[ts, max_usage]],
                'type': 'gauge',
                'tags': self._tags
            })
            limit = lim.get_limit()
            if limit is not None:
                series.append({
                    'metric': '%s.limit' % mname,
                    'points': [[ts, limit]],
                    'type': 'gauge',
                    'tags': self._tags
                })
        logger.info('POSTing %d metrics to datadog', len(series))
        data = {'series': series}
        encoded = json.dumps(data).encode('utf-8')
        url = 'https://api.datadoghq.com/api/v1/series' \
              '?api_key=%s' % self._api_key
        resp = self._http.request(
            'POST', url,
            headers={'Content-type': 'application/json'},
            body=encoded
        )
        if resp.status < 300:
            logger.debug(
                'Successfully POSTed to Datadog; HTTP %d: %s',
                resp.status, resp.data
            )
            return
        raise RuntimeError(
            'ERROR sending metrics to Datadog; API responded HTTP %d: %s' % (
                resp.status, resp.data
            )
        )



"""
awslimitchecker/alerts/pagerdutyv1.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2019 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import os
import logging
import urllib3
import json

from .base import AlertProvider
from awslimitchecker.utils import issue_string_tuple

logger = logging.getLogger(__name__)


class PagerDutyV1(AlertProvider):
    """
    Alert provider to send notifications to PagerDuty via the Events API V1.
    """

    pd_url = 'https://events.pagerduty.com/generic/2010-04-15/create_event.json'

    def __init__(
        self, region_name, account_alias=None, critical_service_key=None,
        warning_service_key=None, incident_key=None
    ):
        """
        Initialize PagerDutyV1 alert provider.

        :param region_name: the name of the region we're connected to
        :type region_name: str
        :param account_alias: Optional; an alias for the account that
          awslimitchecker is currently running against, to use in the default
          incident_key and description.
        :param critical_service_key: **Required**; the PagerDuty Integration
          Key for sending Critical events. Can also be specified via the
          ``PAGERDUTY_SERVICE_KEY_CRIT`` environment variable.
        :type critical_service_key: str
        :param warning_service_key: **Required**; the PagerDuty Integration
          Key for sending Warning events. Can also be specified via the
          ``PAGERDUTY_SERVICE_KEY_WARN`` environment variable. If omitted,
          alerts will not be sent for warnings.
        :type warning_service_key: str
        :param incident_key: Optional; the PagerDuty incident/routing key to
          use, for de-duplication and resolving alerts. This string will have
          any occurrences of ``{account_alias}`` replaced with the account
          alias (or an empty string, if not specified) and any occurrences of
          ``{region_name}`` replaced with the current region name. If not
          specified, this will default to
          ``awslimitchecker-{account_alias}-{region_name}``.
        :type incident_key: str
        """
        super(PagerDutyV1, self).__init__(region_name)
        self._service_key_crit = os.environ.get(
            'PAGERDUTY_SERVICE_KEY_CRIT', None
        )
        if critical_service_key is not None:
            self._service_key_crit = critical_service_key
        if self._service_key_crit is None:
            raise RuntimeError(
                'ERROR: PagerDutyV1 alert provider requires '
                'critical_service_key parameter or PAGERDUTY_SERVICE_KEY_CRIT '
                'environment variable.'
            )
        self._service_key_warn = os.environ.get(
            'PAGERDUTY_SERVICE_KEY_WARN', None
        )
        if warning_service_key is not None:
            self._service_key_warn = warning_service_key
        self._account_alias = account_alias
        if incident_key is None:
            incident_key = 'awslimitchecker-{account_alias}-{region_name}'
        self._incident_key = incident_key.format(
            account_alias='' if self._account_alias is None
            else self._account_alias,
            region_name=self._region_name
        )

    def _send_event(self, service_key, payload):
        """
        Send an event to PagerDuty.

        :param service_key: service key to send to
        :type service_key: str
        :param payload: data to send with event
        :type payload: dict
        """
        payload['service_key'] = service_key
        http = urllib3.PoolManager()
        logger.info(
            'POSTing to PagerDuty Events API (%s): %s', self.pd_url, payload
        )
        encoded = json.dumps(payload, sort_keys=True).encode('utf-8')
        resp = http.request(
            'POST', self.pd_url,
            headers={'Content-type': 'application/json'},
            body=encoded
        )
        if resp.status == 200:
            logger.debug(
                'Successfully POSTed to PagerDuty; HTTP %d: %s',
                resp.status, resp.data
            )
            return
        raise RuntimeError(
            'ERROR creating PagerDuty Event; API responded HTTP %d: %s' % (
                resp.status, resp.data
            )
        )

    def _event_dict(self):
        """
        Return a skeleton dictionary for the PagerDuty V1 Event.

        :return: skeleton of Event
        :rtype: dict
        """
        d = {
            'incident_key': self._incident_key,
            'details': {
                'region': self._region_name
            },
            'client': 'awslimitchecker'
        }
        if self._account_alias is not None:
            d['details']['account_alias'] = self._account_alias
        return d

    def on_success(self, duration=None):
        """
        Method called when no thresholds were breached, and run completed
        successfully. Should resolve any open incidents (if the service supports
        that functionality) or else simply return.

        :param duration: duration of the usage/threshold checking run
        :type duration: float
        """
        data = self._event_dict()
        data['event_type'] = 'resolve'
        data['description'] = 'awslimitchecker in '
        if self._account_alias is not None:
            data['description'] += self._account_alias + ' '
        data['description'] += self._region_name + ' found no problems'
        if duration:
            data['description'] += '; run completed in %.2f seconds' % duration
            data['details']['duration_seconds'] = duration
        self._send_event(self._service_key_crit, data)
        if self._service_key_warn is not None:
            self._send_event(self._service_key_warn, data)

    def _problems_dict(self, problems):
        """
        Make a dict of problems suitable for inclusion in Event details.

        :param problems: dict of service name to nested dict of limit name to
          limit, same format as the return value of
          :py:meth:`~.AwsLimitChecker.check_thresholds`. ``None`` if ``exc`` is
          specified.
        :type problems: dict
        :return: problems summary suitable for Event details
        :rtype: dict
        """
        res = {}
        w_count = 0
        c_count = 0
        for svc in sorted(problems.keys()):
            for lim_name in sorted(problems[svc].keys()):
                limit = problems[svc][lim_name]
                warns = limit.get_warnings()
                w_count += len(warns)
                crits = limit.get_criticals()
                c_count += len(crits)
                _, v = issue_string_tuple(
                    svc, limit, crits, warns, colorize=False
                )
                if svc not in res:
                    res[svc] = {}
                res[svc][lim_name] = v
        return w_count, c_count, res

    def on_critical(self, problems, problem_str, exc=None, duration=None):
        """
        Method called when the run encountered errors, or at least one critical
        threshold was met or crossed.

        :param problems: dict of service name to nested dict of limit name to
          limit, same format as the return value of
          :py:meth:`~.AwsLimitChecker.check_thresholds`. ``None`` if ``exc`` is
          specified.
        :type problems: dict or None
        :param problem_str: String representation of ``problems``, as displayed
          in ``awslimitchecker`` command line output. ``None`` if ``exc`` is
          specified.
        :type problem_str: str or None
        :param exc: Exception object that was raised during the run (optional)
        :type exc: Exception
        :param duration: duration of the run
        :type duration: float
        """
        data = self._event_dict()
        data['event_type'] = 'trigger'
        data['description'] = 'awslimitchecker in '
        if self._account_alias is not None:
            data['description'] += self._account_alias + ' '
        data['description'] += self._region_name
        if duration:
            data['description'] += ' ran in %.2f seconds and' % duration
            data['details']['duration_seconds'] = duration
        if exc is not None:
            data['description'] += ' failed with an exception:' \
                                   ' %s' % exc.__repr__()
            data['details']['exception'] = exc.__repr__()
        else:
            w_count, c_count, pdict = self._problems_dict(problems)
            data['description'] += ' crossed %d CRITICAL thresholds' % c_count
            if w_count > 0:
                data['description'] += ' and %d WARNING thresholds' % w_count
            data['details']['limits'] = pdict
        self._send_event(self._service_key_crit, data)

    def on_warning(self, problems, problem_str, duration=None):
        """
        Method called when one or more warning thresholds were crossed, but no
        criticals and the run did not encounter any errors.

        :param problems: dict of service name to nested dict of limit name to
          limit, same format as the return value of
          :py:meth:`~.AwsLimitChecker.check_thresholds`.
        :type problems: dict or None
        :param problem_str: String representation of ``problems``, as displayed
          in ``awslimitchecker`` command line output.
        :type problem_str: str or None
        :param duration: duration of the run
        :type duration: float
        """
        data = self._event_dict()
        data['event_type'] = 'trigger'
        data['description'] = 'awslimitchecker in '
        if self._account_alias is not None:
            data['description'] += self._account_alias + ' '
        data['description'] += self._region_name
        if duration:
            data['description'] += ' ran in %.2f seconds and' % duration
            data['details']['duration_seconds'] = duration
        w_count, _, pdict = self._problems_dict(problems)
        data['description'] += ' crossed %d WARNING thresholds' % w_count
        data['details']['limits'] = pdict
        self._send_event(self._service_key_warn, data)



"""
awslimitchecker/metrics/base.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2019 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from abc import ABCMeta, abstractmethod

logger = logging.getLogger(__name__)


class MetricsProvider(object):

    __metaclass__ = ABCMeta

    def __init__(self, region_name):
        """
        Initialize a MetricsProvider class. This MUST be overridden by
        subclasses. All configuration must be passed as keyword arguments
        to the class constructor (these come from ``--metrics-config`` CLI
        arguments). Any dependency imports must be made in the constructor.
        The constructor should do as much as possible to validate configuration.

        :param region_name: the name of the region we're connected to
        :type region_name: str
        """
        self._region_name = region_name
        self._duration = 0.0
        self._limits = []

    def set_run_duration(self, duration):
        """
        Set the duration for the awslimitchecker run (the time taken to check
        usage against limits).

        :param duration: time taken to check limits
        :type duration: float
        """
        self._duration = duration

    def add_limit(self, limit):
        """
        Cache a given limit for later sending to the metrics store.

        :param limit: a limit to cache
        :type limit: AwsLimit
        """
        self._limits.append(limit)

    @abstractmethod
    def flush(self):
        """
        Flush all metrics to the provider. This is the method that actually
        sends data to your metrics provider/store. It should iterate over
        ``self._limits`` and send metrics for them, as well as for
        ``self._duration``.
        """
        raise NotImplementedError()

    @staticmethod
    def providers_by_name():
        """
        Return a dict of available MetricsProvider subclass names to the class
        objects.

        :return: MetricsProvider class names to classes
        :rtype: dict
        """
        return {x.__name__: x for x in MetricsProvider.__subclasses__()}

    @staticmethod
    def get_provider_by_name(name):
        """
        Get a reference to the provider class with the specified name.

        :param name: name of the MetricsProvider subclass
        :type name: str
        :return: MetricsProvider subclass
        :rtype: ``class``
        :raises: RuntimeError
        """
        try:
            return MetricsProvider.providers_by_name()[name]
        except KeyError:
            raise RuntimeError(
                'ERROR: "%s" is not a valid MetricsProvider class name' % name
            )



"""
awslimitchecker/runner.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

##############################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
##############################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
##############################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
##############################################################################
"""

import sys
import argparse
import logging
import json
import boto3
import time

from .checker import AwsLimitChecker
from .utils import StoreKeyValuePair, dict2cols, issue_string_tuple
from .limit import SOURCE_TA, SOURCE_API, SOURCE_QUOTAS
from .metrics import MetricsProvider
from .alerts import AlertProvider

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger()

# suppress boto3 internal logging below WARNING level
boto3_log = logging.getLogger("boto3")
boto3_log.setLevel(logging.WARNING)
boto3_log.propagate = True

# suppress botocore internal logging below WARNING level
botocore_log = logging.getLogger("botocore")
botocore_log.setLevel(logging.WARNING)
botocore_log.propagate = True


class Runner(object):

    def __init__(self):
        self.colorize = True
        self.checker = None
        self.skip_ta = False
        self.service_name = None
        self.skip_check = []

    def parse_args(self, argv):
        """
        parse arguments/options

        :param argv: argument list to parse, usually ``sys.argv[1:]``
        :type argv: list
        :returns: parsed arguments
        :rtype: :py:class:`argparse.Namespace`
        """
        desc = 'Report on AWS service limits and usage via boto3, optionally ' \
               'warn about any services with usage nearing or exceeding their' \
               ' limits. For further help, see ' \
               '<http://awslimitchecker.readthedocs.org/>'
        # ###### IMPORTANT license notice ##########
        # Pursuant to Sections 5(b) and 13 of the GNU Affero General Public
        # License, version 3, this notice MUST NOT be removed, and MUST be
        # displayed to ALL USERS of this software, even if they interact with
        # it remotely over a network.
        #
        # See the "Development" section of the awslimitchecker documentation
        # (docs/source/development.rst or
        # <http://awslimitchecker.readthedocs.org/en/latest/development.html> )
        # for further information.
        # ###### IMPORTANT license notice ##########
        epilog = 'awslimitchecker is AGPLv3-licensed Free Software. Anyone ' \
                 'using this program, even remotely over a network, is ' \
                 'entitled to a copy of the source code. Use `--version` for ' \
                 'information on the source code location.'
        p = argparse.ArgumentParser(description=desc, epilog=epilog)
        p.add_argument('-S', '--service', action='store', nargs='*',
                       help='perform action for only the specified service name'
                            '; see -s|--list-services for valid names')
        p.add_argument('--skip-service', action='append', default=[],
                       dest='skip_service',
                       help='avoid performing actions for the specified service'
                            ' name; see -s|--list-services for valid names')
        p.add_argument('--skip-check', action='append', default=[],
                       dest='skip_check',
                       help='avoid performing actions for the specified check'
                            ' name')
        p.add_argument('-s', '--list-services', action='store_true',
                       default=False,
                       help='print a list of all AWS service types that '
                            'awslimitchecker knows how to check')
        p.add_argument('-l', '--list-limits', action='store_true',
                       default=False,
                       help='print all AWS effective limits in "service_name/'
                       'limit_name" format')
        p.add_argument('--list-defaults', action='store_true', default=False,
                       help='print all AWS default limits in "service_name/'
                       'limit_name" format')
        p.add_argument('-L', '--limit', action=StoreKeyValuePair,
                       help='override a single AWS limit, specified in '
                       '"service_name/limit_name=value" format; can be '
                       'specified multiple times.')
        p.add_argument('--limit-override-json', action='store', type=str,
                       default=None,
                       help='Absolute or relative path, or s3:// URL, to a '
                            'JSON file specifying limit overrides. See docs '
                            'for expected format.')
        p.add_argument('--threshold-override-json', action='store', type=str,
                       default=None,
                       help='Absolute or relative path, or s3:// URL, to a '
                            'JSON file specifying threshold overrides. See '
                            'docs for expected format.')
        p.add_argument('-u', '--show-usage', action='store_true',
                       default=False,
                       help='find and print the current usage of all AWS '
                       'services with known limits')
        p.add_argument('--iam-policy', action='store_true',
                       default=False,
                       help='output a JSON serialized IAM Policy '
                       'listing the required permissions for '
                       'awslimitchecker to run correctly.')
        p.add_argument('-W', '--warning-threshold', action='store',
                       type=int, default=80,
                       help='default warning threshold (percentage of '
                       'limit); default: 80')
        p.add_argument('-C', '--critical-threshold', action='store',
                       type=int, default=99,
                       help='default critical threshold (percentage of '
                       'limit); default: 99')
        p.add_argument('-P', '--profile', action='store', dest='profile_name',
                       type=str, default=None,
                       help='Name of profile in the AWS cross-sdk credentials '
                            'file to use credentials from; similar to the '
                            'corresponding awscli option')
        p.add_argument('-A', '--sts-account-id', action='store',
                       type=str, default=None,
                       help='for use with STS, the Account ID of the '
                       'destination account (account to assume a role in)')
        p.add_argument('-R', '--sts-account-role', action='store',
                       type=str, default=None,
                       help='for use with STS, the name of the IAM role to '
                       'assume')
        p.add_argument('-E', '--external-id', action='store', type=str,
                       default=None, help='External ID to use when assuming '
                       'a role via STS')
        p.add_argument('-M', '--mfa-serial-number', action='store', type=str,
                       default=None, help='MFA Serial Number to use when '
                       'assuming a role via STS')
        p.add_argument('-T', '--mfa-token', action='store', type=str,
                       default=None, help='MFA Token to use when assuming '
                       'a role via STS')
        p.add_argument('-r', '--region', action='store',
                       type=str, default=None,
                       help='AWS region name to connect to; required for STS')
        p.add_argument('--role-partition', action='store', type=str,
                       default='aws',
                       help='AWS partition name to use for account_role when '
                            'connecting via STS; see documentation for more '
                            'information (default: "aws")')
        p.add_argument('--ta-api-region', action='store', type=str,
                       default='us-east-1',
                       help='Region to use for Trusted Advisor / Support API'
                            ' (default: us-east-1)')
        p.add_argument('--skip-ta', action='store_true', default=False,
                       help='do not attempt to pull *any* information on limits'
                       ' from Trusted Advisor')
        p.add_argument('--skip-quotas', action='store_true', default=False,
                       help='Do not attempt to connect to Service Quotas '
                            'service or use its data for current limits')
        g = p.add_mutually_exclusive_group()
        g.add_argument('--ta-refresh-wait', dest='ta_refresh_wait',
                       action='store_true', default=False,
                       help='If applicable, refresh all Trusted Advisor '
                            'limit-related checks, and wait for the refresh to'
                            ' complete before continuing.')
        g.add_argument('--ta-refresh-trigger', dest='ta_refresh_trigger',
                       action='store_true', default=False,
                       help='If applicable, trigger refreshes for all Trusted '
                            'Advisor limit-related checks, but do not wait for '
                            'them to finish refreshing; trigger the refresh '
                            'and continue on (useful to ensure checks are '
                            'refreshed before the next scheduled run).')
        g.add_argument('--ta-refresh-older', dest='ta_refresh_older',
                       action='store', type=int, default=None,
                       help='If applicable, trigger refreshes for all Trusted '
                            'Advisor limit-related checks with results more '
                            'than this number of seconds old. Wait for the '
                            'refresh to complete before continuing.')
        p.add_argument('--ta-refresh-timeout', dest='ta_refresh_timeout',
                       type=int, action='store', default=None,
                       help='If waiting for TA checks to refresh, wait up to '
                            'this number of seconds before continuing on '
                            'anyway.')
        p.add_argument('--no-color', action='store_true', default=False,
                       help='do not colorize output')
        p.add_argument('--no-check-version', action='store_false', default=True,
                       dest='check_version',
                       help='do not check latest version at startup')
        p.add_argument('-v', '--verbose', dest='verbose', action='count',
                       default=0,
                       help='verbose output. specify twice for debug-level '
                       'output.')
        p.add_argument('-V', '--version', dest='version', action='store_true',
                       default=False,
                       help='print version number and exit.')
        p.add_argument('--list-metrics-providers',
                       dest='list_metrics_providers',
                       action='store_true', default=False,
                       help='List available metrics providers and exit')
        p.add_argument('--metrics-provider', dest='metrics_provider', type=str,
                       action='store', default=None,
                       help='Metrics provider class name, to enable sending '
                            'metrics')
        p.add_argument('--metrics-config', action=StoreKeyValuePair,
                       dest='metrics_config',
                       help='Specify key/value parameters for the metrics '
                            'provider constructor. See documentation for '
                            'further information.')
        p.add_argument('--list-alert-providers',
                       dest='list_alert_providers',
                       action='store_true', default=False,
                       help='List available alert providers and exit')
        p.add_argument('--alert-provider', dest='alert_provider', type=str,
                       action='store', default=None,
                       help='Alert provider class name, to enable sending '
                            'notifications')
        p.add_argument('--alert-config', action=StoreKeyValuePair,
                       dest='alert_config',
                       help='Specify key/value parameters for the alert '
                            'provider constructor. See documentation for '
                            'further information.')
        args = p.parse_args(argv)
        args.ta_refresh_mode = None
        if args.ta_refresh_wait:
            args.ta_refresh_mode = 'wait'
        elif args.ta_refresh_trigger:
            args.ta_refresh_mode = 'trigger'
        elif args.ta_refresh_older is not None:
            args.ta_refresh_mode = args.ta_refresh_older
        return args

    def list_services(self):
        for x in sorted(self.checker.get_service_names()):
            print(x)

    def list_limits(self):
        limits = self.checker.get_limits(
            use_ta=(not self.skip_ta),
            service=self.service_name)
        data = {}
        for svc in sorted(limits.keys()):
            for lim in sorted(limits[svc].keys()):
                src_str = ''
                if limits[svc][lim].get_limit_source() == SOURCE_API:
                    src_str = ' (API)'
                if limits[svc][lim].get_limit_source() == SOURCE_TA:
                    src_str = ' (TA)'
                if limits[svc][lim].get_limit_source() == SOURCE_QUOTAS:
                    src_str = ' (Quotas)'
                if limits[svc][lim].has_resource_limits():
                    for usage in limits[svc][lim].get_current_usage():
                        id = "{s}/{l}/{r}".format(s=svc, l=lim,
                                                  r=usage.resource_id)
                        data[id] = '{v} (API)'.format(v=usage.get_maximum())
                else:
                    data["{s}/{l}".format(s=svc, l=lim)] = '{v}{t}'.format(
                        v=limits[svc][lim].get_limit(),
                        t=src_str)
        print(dict2cols(data))

    def list_defaults(self):
        limits = self.checker.get_limits(service=self.service_name)
        data = {}
        for svc in sorted(limits.keys()):
            for lim in sorted(limits[svc].keys()):
                data["{s}/{l}".format(s=svc, l=lim)] = '{v}'.format(
                    v=limits[svc][lim].default_limit)
        print(dict2cols(data))

    def iam_policy(self):
        policy = self.checker.get_required_iam_policy()
        print(json.dumps(policy, sort_keys=True, indent=2))

    def show_usage(self):
        self.checker.find_usage(
            service=self.service_name, use_ta=(not self.skip_ta))
        limits = self.checker.get_limits(
            service=self.service_name, use_ta=(not self.skip_ta))
        data = {}
        for svc in sorted(limits.keys()):
            for lim in sorted(limits[svc].keys()):
                data["{s}/{l}".format(s=svc, l=lim)] = '{v}'.format(
                    v=limits[svc][lim].get_current_usage_str())
        print(dict2cols(data))

    def check_thresholds(self, metrics=None):
        have_warn = False
        have_crit = False
        problems = self.checker.check_thresholds(
            use_ta=(not self.skip_ta),
            service=self.service_name
        )
        if metrics:
            for svc, svc_limits in sorted(self.checker.get_limits().items()):
                if self.service_name and svc not in self.service_name:
                    continue
                for _, limit in sorted(svc_limits.items()):
                    metrics.add_limit(limit)
        columns = {}
        for svc in sorted(problems.keys()):
            for lim_name in sorted(problems[svc].keys()):
                check_name = "{svc}/{limit}".format(
                    svc=svc,
                    limit=lim_name,
                )
                if check_name in self.skip_check:
                    continue
                limit = problems[svc][lim_name]
                warns = limit.get_warnings()
                crits = limit.get_criticals()
                if len(crits) > 0:
                    have_crit = True
                if len(warns) > 0:
                    have_warn = True
                k, v = issue_string_tuple(
                    svc, limit, crits, warns, colorize=self.colorize
                )
                columns[k] = v
        d2c = dict2cols(columns)
        print(d2c)
        # might as well use the Nagios exit codes,
        # even though our output doesn't work for that
        if have_crit:
            return 2, problems, d2c
        if have_warn:
            return 1, problems, d2c
        return 0, problems, d2c

    def set_limit_overrides(self, overrides):
        for key in sorted(overrides.keys()):
            if key.count('/') != 1:
                raise ValueError("Limit names must be in 'service/limit' "
                                 "format; {k} is invalid.".format(k=key))
            svc, limit = key.split('/')
            self.checker.set_limit_override(svc, limit, int(overrides[key]))

    def load_json(self, path):
        """Load JSON from either a local file or S3"""
        if path.startswith('s3://'):
            parsed = urlparse(path)
            s3key = parsed.path.lstrip('/')
            logger.debug(
                'Reading JSON from S3 bucket "%s" key "%s"',
                parsed.netloc, s3key
            )
            client = boto3.client('s3')
            resp = client.get_object(Bucket=parsed.netloc, Key=s3key)
            data = resp['Body'].read()
        else:
            logger.debug('Reading JSON from: %s', path)
            with open(path, 'r') as fh:
                data = fh.read()
        if isinstance(data, type(b'')):
            data = data.decode()
        return json.loads(data)

    def set_limit_overrides_from_json(self, path):
        j = self.load_json(path)
        logger.debug('Limit overrides: %s', j)
        self.checker.set_limit_overrides(j)
        logger.debug('Done setting limit overrides from JSON.')

    def set_threshold_overrides_from_json(self, path):
        j = self.load_json(path)
        logger.debug('Threshold overrides: %s', j)
        self.checker.set_threshold_overrides(j)
        logger.debug('Done setting threshold overrides from JSON.')

    def console_entry_point(self):
        args = self.parse_args(sys.argv[1:])
        self.service_name = args.service
        if args.verbose == 1:
            logger.setLevel(logging.INFO)
        elif args.verbose > 1:
            # debug-level logging hacks
            FORMAT = "%(asctime)s [%(levelname)s %(filename)s:%(lineno)s - " \
                     "%(name)s.%(funcName)s() ] %(message)s"
            debug_formatter = logging.Formatter(fmt=FORMAT)
            logger.handlers[0].setFormatter(debug_formatter)
            logger.setLevel(logging.DEBUG)

        if args.no_color:
            self.colorize = False

        if args.skip_ta:
            self.skip_ta = True

        # the rest of these actually use the checker
        self.checker = AwsLimitChecker(
            warning_threshold=args.warning_threshold,
            critical_threshold=args.critical_threshold,
            profile_name=args.profile_name,
            account_id=args.sts_account_id,
            account_role=args.sts_account_role,
            region=args.region,
            external_id=args.external_id,
            mfa_serial_number=args.mfa_serial_number,
            mfa_token=args.mfa_token,
            ta_refresh_mode=args.ta_refresh_mode,
            ta_refresh_timeout=args.ta_refresh_timeout,
            check_version=args.check_version,
            role_partition=args.role_partition,
            ta_api_region=args.ta_api_region,
            skip_quotas=args.skip_quotas
        )

        if args.version:
            print('awslimitchecker {v} (see <{s}> for source code)'.format(
                s=self.checker.get_project_url(),
                v=self.checker.get_version()
            ))
            raise SystemExit(0)

        if len(args.skip_service) > 0:
            self.checker.remove_services(args.skip_service)

        if len(args.skip_check) > 0:
            for check in args.skip_check:
                self.skip_check.append(check)

        if args.limit_override_json is not None:
            self.set_limit_overrides_from_json(args.limit_override_json)

        if args.threshold_override_json is not None:
            self.set_threshold_overrides_from_json(
                args.threshold_override_json
            )

        if len(args.limit) > 0:
            self.set_limit_overrides(args.limit)

        if args.list_services:
            self.list_services()
            raise SystemExit(0)

        if args.list_defaults:
            self.list_defaults()
            raise SystemExit(0)

        if args.list_limits:
            self.list_limits()
            raise SystemExit(0)

        if args.iam_policy:
            self.iam_policy()
            raise SystemExit(0)

        if args.show_usage:
            self.show_usage()
            raise SystemExit(0)

        if args.list_metrics_providers:
            print('Available metrics providers:')
            for p in sorted(MetricsProvider.providers_by_name().keys()):
                print(p)
            raise SystemExit(0)

        if args.list_alert_providers:
            print('Available alert providers:')
            for p in sorted(AlertProvider.providers_by_name().keys()):
                print(p)
            raise SystemExit(0)

        # else check
        alerter = None
        if args.alert_provider:
            alerter = AlertProvider.get_provider_by_name(
                args.alert_provider
            )(self.checker.region_name, **args.alert_config)
        start_time = time.time()
        try:
            metrics = None
            if args.metrics_provider:
                metrics = MetricsProvider.get_provider_by_name(
                    args.metrics_provider
                )(self.checker.region_name, **args.metrics_config)
            res, problems, problem_str = self.check_thresholds(metrics)
            duration = time.time() - start_time
            logger.info('Finished checking limits in %s seconds', duration)
            if metrics:
                metrics.set_run_duration(duration)
                metrics.flush()
        except Exception as ex:
            if alerter:
                alerter.on_critical(
                    None, None, exc=ex, duration=time.time() - start_time
                )
            raise
        if alerter:
            if res == 2:
                alerter.on_critical(
                    problems, problem_str, duration=time.time() - start_time
                )
            elif res == 1:
                alerter.on_warning(
                    problems, problem_str, duration=time.time() - start_time
                )
            else:
                alerter.on_success(duration=time.time() - start_time)
            # with alert provider, always exit zero
            raise SystemExit(0)
        raise SystemExit(res)


def console_entry_point():
    r = Runner()
    r.console_entry_point()


if __name__ == "__main__":
    console_entry_point()



"""
awslimitchecker/tests/services/test_elb.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
from awslimitchecker.tests.services import result_fixtures
from awslimitchecker.services.elb import _ElbService, allow_count_or_none_units

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import patch, call, Mock, PropertyMock
else:
    from unittest.mock import patch, call, Mock, PropertyMock


pbm = 'awslimitchecker.services.elb'  # patch base path - module
pb = '%s._ElbService' % pbm  # patch base path


class TestAllowCountOrNoneUnits:

    def test_none(self):
        assert allow_count_or_none_units(10, 'None', 'Count') == 10

    def test_count(self):
        assert allow_count_or_none_units(10, 'Count', 'Count') == 10

    def test_other(self):
        assert allow_count_or_none_units(10, 'Other', 'Count') is None


class Test_ElbService(object):

    def test_init(self):
        """test __init__()"""
        cls = _ElbService(21, 43, {}, None)
        assert cls.service_name == 'ELB'
        assert cls.conn is None
        assert cls.warning_threshold == 21
        assert cls.critical_threshold == 43

    def test_get_limits(self):
        cls = _ElbService(21, 43, {}, None)
        cls.limits = {}
        res = cls.get_limits()
        assert sorted(res.keys()) == sorted([
            'Application load balancers',
            'Certificates per application load balancer',
            'Classic load balancers',
            'Listeners per application load balancer',
            'Listeners per load balancer',
            'Listeners per network load balancer',
            'Network load balancers',
            'Registered instances per load balancer',
            'Rules per application load balancer',
            'Target groups'
        ])
        for name, limit in res.items():
            assert limit.service == cls
            assert limit.def_warning_threshold == 21
            assert limit.def_critical_threshold == 43
            if name in [
                'Application load balancers',
                'Classic load balancers',
            ]:
                assert limit.quotas_unit_converter == allow_count_or_none_units
            else:
                assert limit.quotas_unit_converter is None

    def test_get_limits_again(self):
        """test that existing limits dict is returned on subsequent calls"""
        mock_limits = Mock()
        cls = _ElbService(21, 43, {}, None)
        cls.limits = mock_limits
        res = cls.get_limits()
        assert res == mock_limits

    def test_update_limits_from_api(self):
        r1 = result_fixtures.ELB.test_update_limits_elb
        r2 = result_fixtures.ELB.test_update_limits_alb

        mock_conn = Mock()
        mock_conn.describe_account_limits.return_value = r1

        with patch('%s.connect' % pb) as mock_connect:
            with patch('%s.client' % pbm) as mock_client:
                m_cli = mock_client.return_value
                m_cli._client_config.region_name = PropertyMock(
                    return_value='rname'
                )
                m_cli.describe_account_limits.return_value = r2
                cls = _ElbService(21, 43, {}, None)
                cls.conn = mock_conn
                cls._boto3_connection_kwargs = {'foo': 'bar', 'baz': 'blam'}
                cls.get_limits()
                cls._update_limits_from_api()
        assert mock_connect.mock_calls == [call()]
        assert mock_conn.mock_calls == [call.describe_account_limits()]
        assert mock_client.mock_calls == [
            call('elbv2', foo='bar', baz='blam'),
            call().describe_account_limits()
        ]
        assert cls.limits['Classic load balancers'].api_limit == 3
        assert cls.limits['Application load balancers'].api_limit == 6
        assert cls.limits['Listeners per load balancer'].api_limit == 5
        assert cls.limits[
            'Registered instances per load balancer'].api_limit == 1800
        assert cls.limits['Target groups'].api_limit == 7
        assert cls.limits[
            'Listeners per application load balancer'].api_limit == 9
        assert cls.limits['Rules per application load balancer'].api_limit == 10
        assert cls.limits[
            'Listeners per network load balancer'].api_limit == 100
        assert cls.limits['Network load balancers'].api_limit == 40

    def test_find_usage(self):
        with patch('%s._find_usage_elbv1' % pb, autospec=True) as mock_v1:
            with patch('%s._find_usage_elbv2' % pb, autospec=True) as mock_v2:
                mock_v1.return_value = 3
                mock_v2.return_value = 5
                cls = _ElbService(21, 43, {}, None)
                assert cls._have_usage is False
                cls.find_usage()
        assert cls._have_usage is True
        assert mock_v1.mock_calls == [call(cls)]
        assert mock_v2.mock_calls == [call(cls)]
        assert len(cls.limits[
            'Classic load balancers'].get_current_usage()) == 1
        assert len(cls.limits[
            'Application load balancers'].get_current_usage()) == 1
        assert cls.limits[
            'Classic load balancers'].get_current_usage()[0].get_value() == 3
        assert cls.limits['Application load balancers'
                          ].get_current_usage()[0].get_value() == 5

    def test_find_usage_elbv1(self):
        mock_conn = Mock()

        return_value = result_fixtures.ELB.test_find_usage

        with patch('%s.connect' % pb) as mock_connect:
            with patch('%s.paginate_dict' % pbm) as mock_paginate:
                mock_paginate.return_value = return_value
                cls = _ElbService(21, 43, {}, None)
                cls.conn = mock_conn
                res = cls._find_usage_elbv1()
        assert res == 4
        assert mock_connect.mock_calls == [call()]
        assert mock_conn.mock_calls == []
        assert mock_paginate.mock_calls == [
            call(
                mock_conn.describe_load_balancers,
                alc_marker_path=['NextMarker'],
                alc_data_path=['LoadBalancerDescriptions'],
                alc_marker_param='Marker'
            )
        ]
        entries = sorted(cls.limits[
            'Listeners per load balancer'].get_current_usage())
        assert len(entries) == 4
        assert entries[0].resource_id == 'elb-1'
        assert entries[0].get_value() == 1
        assert entries[1].resource_id == 'elb-2'
        assert entries[1].get_value() == 2
        assert entries[2].resource_id == 'elb-3'
        assert entries[2].get_value() == 3
        assert entries[3].resource_id == 'elb-4'
        assert entries[3].get_value() == 6
        entries = sorted(cls.limits[
            'Registered instances per load balancer'].get_current_usage())
        assert len(entries) == 4
        assert entries[0].resource_id == 'elb-1'
        assert entries[0].get_value() == 0
        assert entries[1].resource_id == 'elb-3'
        assert entries[1].get_value() == 1
        assert entries[2].resource_id == 'elb-4'
        assert entries[2].get_value() == 2
        assert entries[3].resource_id == 'elb-2'
        assert entries[3].get_value() == 4

    def test_find_usage_elbv2(self):
        lbs_res = result_fixtures.ELB.test_find_usage_elbv2_elbs
        tgs_res = result_fixtures.ELB.test_find_usage_elbv2_target_groups

        with patch('%s.connect' % pb) as mock_connect:
            with patch('%s.client' % pbm) as mock_client:
                mock_client.return_value._client_config.region_name = \
                    PropertyMock(return_value='rname')
                with patch('%s.paginate_dict' % pbm) as mock_paginate:
                    with patch(
                        '%s._update_usage_for_alb' % pb, autospec=True
                    ) as mock_u:
                        with patch(
                            '%s.Config' % pbm, autospec=True
                        ) as mock_conf:
                            mock_paginate.side_effect = [
                                tgs_res,
                                lbs_res
                            ]
                            cls = _ElbService(21, 43, {}, None)
                            cls._boto3_connection_kwargs = {
                                'foo': 'bar',
                                'baz': 'blam'
                            }
                            res = cls._find_usage_elbv2()
        assert res == 2
        assert mock_conf.mock_calls == [
            call(retries={'max_attempts': 12})
        ]
        assert mock_connect.mock_calls == []
        assert mock_client.mock_calls == [
            call('elbv2', foo='bar', baz='blam', config=mock_conf.return_value),
        ]
        assert mock_paginate.mock_calls == [
            call(
                mock_client.return_value.describe_target_groups,
                alc_marker_path=['NextMarker'],
                alc_data_path=['TargetGroups'],
                alc_marker_param='Marker'
            ),
            call(
                mock_client.return_value.describe_load_balancers,
                alc_marker_path=['NextMarker'],
                alc_data_path=['LoadBalancers'],
                alc_marker_param='Marker'
            )
        ]
        assert mock_u.mock_calls == [
            call(cls, mock_client.return_value, 'lb-arn1', 'lb1'),
            call(cls, mock_client.return_value, 'lb-arn2', 'lb2')
        ]
        lim = cls.limits['Target groups'].get_current_usage()
        assert len(lim) == 1
        assert lim[0].get_value() == 3
        assert lim[0].aws_type == 'AWS::ElasticLoadBalancingV2::TargetGroup'
        lim = cls.limits['Network load balancers'].get_current_usage()
        assert len(lim) == 1
        assert lim[0].get_value() == 1
        assert lim[0].aws_type == \
            'AWS::ElasticLoadBalancing::NetworkLoadBalancer'

    def test_update_usage_for_alb(self):
        conn = Mock()
        with patch('%s.paginate_dict' % pbm) as mock_paginate:
            mock_paginate.side_effect = [
                result_fixtures.ELB.test_usage_alb_listeners,
                result_fixtures.ELB.test_usage_alb_rules[0],
                result_fixtures.ELB.test_usage_alb_rules[1],
                result_fixtures.ELB.test_usage_alb_rules[2]
            ]
            cls = _ElbService(21, 43, {}, None)
            cls._update_usage_for_alb(conn, 'myarn', 'albname')
        assert mock_paginate.mock_calls == [
            call(
                conn.describe_listeners,
                LoadBalancerArn='myarn',
                alc_marker_path=['NextMarker'],
                alc_data_path=['Listeners'],
                alc_marker_param='Marker'
            ),
            call(
                conn.describe_rules,
                ListenerArn='listener1',
                alc_marker_path=['NextMarker'],
                alc_data_path=['Rules'],
                alc_marker_param='Marker'
            ),
            call(
                conn.describe_rules,
                ListenerArn='listener2',
                alc_marker_path=['NextMarker'],
                alc_data_path=['Rules'],
                alc_marker_param='Marker'
            ),
            call(
                conn.describe_rules,
                ListenerArn='listener3',
                alc_marker_path=['NextMarker'],
                alc_data_path=['Rules'],
                alc_marker_param='Marker'
            )
        ]
        lim = cls.limits[
            'Listeners per application load balancer'].get_current_usage()
        assert len(lim) == 1
        assert lim[0].get_value() == 3
        assert lim[0].aws_type == 'AWS::ElasticLoadBalancingV2::LoadBalancer'
        assert lim[0].resource_id == 'albname'
        r = cls.limits[
            'Rules per application load balancer'].get_current_usage()
        assert len(r) == 1
        assert r[0].get_value() == 7
        assert r[0].aws_type == 'AWS::ElasticLoadBalancingV2::LoadBalancer'
        assert r[0].resource_id == 'albname'
        certs = cls.limits[
            'Certificates per application load balancer'
        ].get_current_usage()
        assert len(certs) == 1
        assert certs[0].get_value() == 3
        assert certs[0].aws_type == 'AWS::ElasticLoadBalancingV2::LoadBalancer'
        assert certs[0].resource_id == 'albname'

    def test_update_usage_for_nlb(self):
        conn = Mock()
        with patch('%s.paginate_dict' % pbm) as mock_paginate:
            mock_paginate.side_effect = [
                result_fixtures.ELB.test_usage_nlb_listeners
            ]
            cls = _ElbService(21, 43, {}, None)
            cls._update_usage_for_nlb(conn, 'mynarn', 'nlbname')
        assert mock_paginate.mock_calls == [
            call(
                conn.describe_listeners,
                LoadBalancerArn='mynarn',
                alc_marker_path=['NextMarker'],
                alc_data_path=['Listeners'],
                alc_marker_param='Marker'
            )
        ]
        lim = cls.limits[
            'Listeners per network load balancer'].get_current_usage()
        assert len(lim) == 1
        assert lim[0].get_value() == 2
        assert lim[0].aws_type == \
            'AWS::ElasticLoadBalancingV2::NetworkLoadBalancer'
        assert lim[0].resource_id == 'nlbname'

    def test_required_iam_permissions(self):
        cls = _ElbService(21, 43, {}, None)
        assert cls.required_iam_permissions() == [
            "elasticloadbalancing:DescribeLoadBalancers",
            "elasticloadbalancing:DescribeAccountLimits",
            "elasticloadbalancing:DescribeListeners",
            "elasticloadbalancing:DescribeTargetGroups",
            "elasticloadbalancing:DescribeRules"
        ]



"""
awslimitchecker/services/elb.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import abc  # noqa
import logging
from boto3 import client
from botocore.config import Config

from .base import _AwsService
from ..limit import AwsLimit
from ..utils import paginate_dict

logger = logging.getLogger(__name__)

#: Override the elbv2 API maximum retry attempts
ELBV2_MAX_RETRY_ATTEMPTS = 12


def allow_count_or_none_units(value, in_unit, out_unit):
    """
    This is a unit converter for Service Quotas; see
    :py:meth:`.ServiceQuotasClient.get_quota_value` for details.

    This is a work-around for
    https://github.com/jantman/awslimitchecker/issues/503 where, sometime
    between 2020-11-02 and 2020-11-09, the quota unit for Application Load
    Balancers per Region and Classic Load Balancers per Region changed, without
    announcement or warning, from "Count" to "None". This converter allows both
    options and treats them identically.
    """
    if in_unit not in ['None', 'Count'] or out_unit != 'Count':
        logger.error(
            'ERROR: cannot convert Service Quotas (A|E)LB limit value from '
            'units of "%s" to units of "%s"', in_unit, out_unit
        )
        return None
    return value


class _ElbService(_AwsService):
    """
    Note that ELB (ELBv1) and ALB (ELBv2) are combined in the same service.
    This is because, per AWS docs, the limit for number of load balancers
    per region is a combination of ELB and ALB.
    """

    service_name = 'ELB'
    api_name = 'elb'
    quotas_service_code = 'elasticloadbalancing'

    def find_usage(self):
        """
        Determine the current usage for each limit of this service,
        and update corresponding Limit via
        :py:meth:`~.AwsLimit._add_current_usage`.
        """
        logger.debug("Checking usage for service %s", self.service_name)
        for lim in self.limits.values():
            lim._reset_usage()
        elb_usage = self._find_usage_elbv1()
        alb_usage = self._find_usage_elbv2()
        logger.debug('ELBs in use: %d, ALBs in use: %d', elb_usage, alb_usage)
        self.limits['Classic load balancers']._add_current_usage(
            elb_usage,
            aws_type='AWS::ElasticLoadBalancing::LoadBalancer',
        )
        self.limits['Application load balancers']._add_current_usage(
            alb_usage,
            aws_type='AWS::ElasticLoadBalancingV2::LoadBalancer',
        )
        self._have_usage = True
        logger.debug("Done checking usage.")

    def _find_usage_elbv1(self):
        """
        Find usage for ELBv1 / Classic ELB and update the appropriate limits.

        :returns: number of Classic ELBs in use
        :rtype: int
        """
        logger.debug("Checking usage for ELBv1")
        self.connect()
        lbs = paginate_dict(
            self.conn.describe_load_balancers,
            alc_marker_path=['NextMarker'],
            alc_data_path=['LoadBalancerDescriptions'],
            alc_marker_param='Marker'
        )
        for lb in lbs['LoadBalancerDescriptions']:
            self.limits['Listeners per load balancer']._add_current_usage(
                len(lb['ListenerDescriptions']),
                aws_type='AWS::ElasticLoadBalancing::LoadBalancer',
                resource_id=lb['LoadBalancerName'],
            )
            self.limits[
                'Registered instances per load balancer'
            ]._add_current_usage(
                len(lb['Instances']),
                aws_type='AWS::ElasticLoadBalancing::LoadBalancer',
                resource_id=lb['LoadBalancerName']
            )
        logger.debug('Done with ELBv1 usage')
        return len(lbs['LoadBalancerDescriptions'])

    def _find_usage_elbv2(self):
        """
        Find usage for ELBv2 / Application LB and update the appropriate limits.

        :returns: number of Application LBs in use
        :rtype: int
        """
        logger.debug('Checking usage for ELBv2')
        conn2 = client(
            'elbv2',
            config=Config(retries={'max_attempts': ELBV2_MAX_RETRY_ATTEMPTS}),
            **self._boto3_connection_kwargs
        )
        logger.debug("Connected to %s in region %s (with max retry attempts "
                     "overridden to %d)", 'elbv2',
                     conn2._client_config.region_name, ELBV2_MAX_RETRY_ATTEMPTS)
        # Target groups
        tgroups = paginate_dict(
            conn2.describe_target_groups,
            alc_marker_path=['NextMarker'],
            alc_data_path=['TargetGroups'],
            alc_marker_param='Marker'
        )
        self.limits['Target groups']._add_current_usage(
            len(tgroups['TargetGroups']),
            aws_type='AWS::ElasticLoadBalancingV2::TargetGroup'
        )
        # ALBs
        lbs = paginate_dict(
            conn2.describe_load_balancers,
            alc_marker_path=['NextMarker'],
            alc_data_path=['LoadBalancers'],
            alc_marker_param='Marker'
        )
        logger.debug(
            'Checking usage for each of %d ALBs', len(lbs['LoadBalancers'])
        )
        alb_count = 0
        nlb_count = 0
        for lb in lbs['LoadBalancers']:
            if lb.get('Type') == 'network':
                nlb_count += 1
            else:
                alb_count += 1
                self._update_usage_for_alb(
                    conn2,
                    lb['LoadBalancerArn'],
                    lb['LoadBalancerName']
                )
        self.limits['Network load balancers']._add_current_usage(
            nlb_count,
            aws_type='AWS::ElasticLoadBalancing::NetworkLoadBalancer'
        )
        logger.debug('Done with ELBv2 usage')
        return alb_count

    def _update_usage_for_alb(self, conn, alb_arn, alb_name):
        """
        Update usage for a single ALB.

        :param conn: elbv2 API connection
        :type conn: :py:class:`ElasticLoadBalancing.Client`
        :param alb_arn: Load Balancer ARN
        :type alb_arn: str
        :param alb_name: Load Balancer Name
        :type alb_name: str
        """
        logger.debug('Updating usage for ALB %s', alb_arn)
        listeners = paginate_dict(
            conn.describe_listeners,
            LoadBalancerArn=alb_arn,
            alc_marker_path=['NextMarker'],
            alc_data_path=['Listeners'],
            alc_marker_param='Marker'
        )['Listeners']
        num_rules = 0
        num_certs = 0
        for l in listeners:
            certs = [
                x for x in l.get('Certificates', [])
                if x.get('IsDefault', False) is False
            ]
            num_certs += len(certs)
            rules = paginate_dict(
                conn.describe_rules,
                ListenerArn=l['ListenerArn'],
                alc_marker_path=['NextMarker'],
                alc_data_path=['Rules'],
                alc_marker_param='Marker'
            )['Rules']
            num_rules += len(rules)
        self.limits[
            'Listeners per application load balancer']._add_current_usage(
            len(listeners),
            aws_type='AWS::ElasticLoadBalancingV2::LoadBalancer',
            resource_id=alb_name,
        )
        self.limits['Rules per application load balancer']._add_current_usage(
            num_rules,
            aws_type='AWS::ElasticLoadBalancingV2::LoadBalancer',
            resource_id=alb_name,
        )
        self.limits[
            'Certificates per application load balancer'
        ]._add_current_usage(
            num_certs,
            aws_type='AWS::ElasticLoadBalancingV2::LoadBalancer',
            resource_id=alb_name
        )

    def _update_usage_for_nlb(self, conn, nlb_arn, nlb_name):
        """
        Update usage for a single NLB.

        :param conn: elbv2 API connection
        :type conn: :py:class:`ElasticLoadBalancing.Client`
        :param nlb_arn: Load Balancer ARN
        :type nlb_arn: str
        :param nlb_name: Load Balancer Name
        :type nlb_name: str
        """
        logger.debug('Updating usage for NLB %s', nlb_arn)
        listeners = paginate_dict(
            conn.describe_listeners,
            LoadBalancerArn=nlb_arn,
            alc_marker_path=['NextMarker'],
            alc_data_path=['Listeners'],
            alc_marker_param='Marker'
        )['Listeners']
        self.limits[
            'Listeners per network load balancer']._add_current_usage(
            len(listeners),
            aws_type='AWS::ElasticLoadBalancingV2::NetworkLoadBalancer',
            resource_id=nlb_name
        )

    def get_limits(self):
        """
        Return all known limits for this service, as a dict of their names
        to :py:class:`~.AwsLimit` objects.

        :returns: dict of limit names to :py:class:`~.AwsLimit` objects
        :rtype: dict
        """
        if self.limits != {}:
            return self.limits
        limits = {}
        # ELBv1 (Classic ELB) limits
        limits['Classic load balancers'] = AwsLimit(
            'Classic load balancers',
            self,
            20,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::ElasticLoadBalancing::LoadBalancer',
            quotas_name='Classic Load Balancers per Region',
            quotas_unit='Count',
            quotas_unit_converter=allow_count_or_none_units
        )
        limits['Listeners per load balancer'] = AwsLimit(
            'Listeners per load balancer',
            self,
            100,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::ElasticLoadBalancing::LoadBalancer',
            limit_subtype='LoadBalancerListener'
        )
        limits['Registered instances per load balancer'] = AwsLimit(
            'Registered instances per load balancer',
            self,
            1000,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::ElasticLoadBalancing::LoadBalancer',
            limit_subtype='Instance'
        )
        # ELBv2 (ALB) limits
        limits['Application load balancers'] = AwsLimit(
            'Application load balancers',
            self,
            20,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::ElasticLoadBalancingV2::LoadBalancer',
            quotas_name='Application Load Balancers per Region',
            quotas_unit='Count',
            quotas_unit_converter=allow_count_or_none_units
        )
        limits['Target groups'] = AwsLimit(
            'Target groups',
            self,
            3000,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::ElasticLoadBalancingV2::LoadBalancer',
            limit_subtype='LoadBalancerTargetGroup'
        )
        limits['Listeners per application load balancer'] = AwsLimit(
            'Listeners per application load balancer',
            self,
            50,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::ElasticLoadBalancingV2::LoadBalancer',
            limit_subtype='LoadBalancerListener'
        )
        limits['Certificates per application load balancer'] = AwsLimit(
            'Certificates per application load balancer',
            self,
            25,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::ElasticLoadBalancingV2::LoadBalancer',
            limit_subtype='Certificate'
        )
        limits['Rules per application load balancer'] = AwsLimit(
            'Rules per application load balancer',
            self,
            100,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::ElasticLoadBalancingV2::LoadBalancer',
            limit_subtype='LoadBalancerRule'
        )
        limits['Network load balancers'] = AwsLimit(
            'Network load balancers',
            self,
            20,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::ElasticLoadBalancing::NetworkLoadBalancer',
        )
        limits['Listeners per network load balancer'] = AwsLimit(
            'Listeners per network load balancer',
            self,
            50,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::ElasticLoadBalancingV2::NetworkLoadBalancer',
            limit_subtype='LoadBalancerListener'
        )
        self.limits = limits
        return limits

    def _update_limits_from_api(self):
        """
        Query ELB's DescribeAccountLimits API action, and update limits
        with the quotas returned. Updates ``self.limits``.
        """
        self.connect()
        logger.debug("Querying ELB DescribeAccountLimits for limits")
        attribs = self.conn.describe_account_limits()
        name_to_limits = {
            'classic-load-balancers': 'Classic load balancers',
            'classic-listeners': 'Listeners per load balancer',
            'classic-registered-instances':
                'Registered instances per load balancer'
        }
        for attrib in attribs['Limits']:
            if int(attrib.get('Max', 0)) == 0:
                continue
            name = attrib.get('Name', 'unknown')
            if name not in name_to_limits:
                continue
            self.limits[name_to_limits[name]]._set_api_limit(int(attrib['Max']))
        # connect to ELBv2 API as well
        self.conn2 = client('elbv2', **self._boto3_connection_kwargs)
        logger.debug("Connected to %s in region %s",
                     'elbv2', self.conn2._client_config.region_name)
        logger.debug("Querying ELBv2 (ALB) DescribeAccountLimits for limits")
        attribs = self.conn2.describe_account_limits()
        name_to_limits = {
            'application-load-balancers': 'Application load balancers',
            'target-groups': 'Target groups',
            'listeners-per-application-load-balancer':
                'Listeners per application load balancer',
            'rules-per-application-load-balancer':
                'Rules per application load balancer',
            'network-load-balancers': 'Network load balancers',
            'listeners-per-network-load-balancer':
                'Listeners per network load balancer'
        }
        for attrib in attribs['Limits']:
            if int(attrib.get('Max', 0)) == 0:
                continue
            name = attrib.get('Name', 'unknown')
            if name not in name_to_limits:
                continue
            self.limits[name_to_limits[name]]._set_api_limit(int(attrib['Max']))
        logger.debug("Done setting limits from API")

    def required_iam_permissions(self):
        """
        Return a list of IAM Actions required for this Service to function
        properly. All Actions will be shown with an Effect of "Allow"
        and a Resource of "*".

        :returns: list of IAM Action strings
        :rtype: list
        """
        return [
            "elasticloadbalancing:DescribeLoadBalancers",
            "elasticloadbalancing:DescribeAccountLimits",
            "elasticloadbalancing:DescribeListeners",
            "elasticloadbalancing:DescribeTargetGroups",
            "elasticloadbalancing:DescribeRules"
        ]



"""
awslimitchecker/tests/metrics/test_datadog.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2019 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
import json
from awslimitchecker.metrics import Datadog
import pytest
from freezegun import freeze_time

if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import Mock, patch, call
else:
    from unittest.mock import Mock, patch, call


pbm = 'awslimitchecker.metrics.datadog'
pb = '%s.Datadog' % pbm


class TestInit(object):

    @patch.dict('os.environ', {}, clear=True)
    def test_happy_path(self):
        mock_http = Mock()
        with patch('%s.urllib3.PoolManager' % pbm, autospec=True) as m_pm:
            m_pm.return_value = mock_http
            with patch('%s._validate_auth' % pb, autospec=True) as m_va:
                cls = Datadog(
                    'foo', api_key='1234', extra_tags='foo,bar,baz:blam'
                )
        assert cls._region_name == 'foo'
        assert cls._duration == 0.0
        assert cls._limits == []
        assert cls._api_key == '1234'
        assert cls._prefix == 'awslimitchecker.'
        assert cls._tags == [
            'region:foo', 'foo', 'bar', 'baz:blam'
        ]
        assert cls._http == mock_http
        assert m_pm.mock_calls == [call()]
        assert m_va.mock_calls == [call(cls, '1234')]

    @patch.dict('os.environ', {'DATADOG_API_KEY': '5678'}, clear=True)
    def test_api_key_env_var(self):
        mock_http = Mock()
        with patch('%s.urllib3.PoolManager' % pbm, autospec=True) as m_pm:
            m_pm.return_value = mock_http
            with patch('%s._validate_auth' % pb, autospec=True) as m_va:
                cls = Datadog(
                    'foo', prefix='myprefix.'
                )
        assert cls._region_name == 'foo'
        assert cls._duration == 0.0
        assert cls._limits == []
        assert cls._api_key == '5678'
        assert cls._prefix == 'myprefix.'
        assert cls._tags == [
            'region:foo'
        ]
        assert cls._http == mock_http
        assert m_pm.mock_calls == [call()]
        assert m_va.mock_calls == [call(cls, '5678')]

    @patch.dict('os.environ', {}, clear=True)
    def test_no_api_key(self):
        mock_http = Mock()
        with patch('%s.urllib3.PoolManager' % pbm, autospec=True) as m_pm:
            m_pm.return_value = mock_http
            with patch('%s._validate_auth' % pb, autospec=True) as m_va:
                with pytest.raises(RuntimeError) as exc:
                    Datadog(
                        'foo', extra_tags='foo,bar,baz:blam', prefix='myprefix.'
                    )
        assert str(exc.value) == 'ERROR: Datadog metrics provider ' \
                                 'requires datadog API key.'
        assert m_pm.mock_calls == []
        assert m_va.mock_calls == []


class DatadogTester(object):

    def setup(self):
        with patch('%s.__init__' % pb) as m_init:
            m_init.return_value = None
            self.cls = Datadog()


class TestValidateAuth(DatadogTester):

    def test_happy_path(self):
        mock_http = Mock()
        mock_resp = Mock(status=200, data=b'{"success": "ok"}')
        mock_http.request.return_value = mock_resp
        self.cls._http = mock_http
        self.cls._validate_auth('1234')
        assert mock_http.mock_calls == [
            call.request(
                'GET',
                'https://api.datadoghq.com/api/v1/validate?api_key=1234'
            )
        ]

    def test_failure(self):
        mock_http = Mock()
        mock_resp = Mock(status=401, data='{"success": "NO"}')
        mock_http.request.return_value = mock_resp
        self.cls._http = mock_http
        with pytest.raises(RuntimeError) as exc:
            self.cls._validate_auth('1234')
        assert str(exc.value) == 'ERROR: Datadog API key validation failed ' \
                                 'with HTTP 401: {"success": "NO"}'
        assert mock_http.mock_calls == [
            call.request(
                'GET',
                'https://api.datadoghq.com/api/v1/validate?api_key=1234'
            )
        ]


class TestNameForMetric(DatadogTester):

    def test_simple(self):
        self.cls._prefix = 'foobar.'
        assert self.cls._name_for_metric(
            'Service Name*', 'limit NAME .'
        ) == 'foobar.service_name_.limit_name_'


class TestFlush(DatadogTester):

    @freeze_time("2016-12-16 10:40:42", tz_offset=0, auto_tick_seconds=6)
    def test_happy_path(self):
        self.cls._prefix = 'prefix.'
        self.cls._tags = ['tag1', 'tag:2']
        self.cls._limits = []
        self.cls._api_key = 'myKey'
        self.cls.set_run_duration(123.45)
        limA = Mock(
            name='limitA', service=Mock(service_name='SVC1')
        )
        type(limA).name = 'limitA'
        limA.get_current_usage.return_value = []
        limA.get_limit.return_value = None
        self.cls.add_limit(limA)
        limB = Mock(
            name='limitB', service=Mock(service_name='SVC1')
        )
        type(limB).name = 'limitB'
        mocku = Mock()
        mocku.get_value.return_value = 6
        limB.get_current_usage.return_value = [mocku]
        limB.get_limit.return_value = 10
        self.cls.add_limit(limB)
        mock_http = Mock()
        mock_resp = Mock(status=200, data='{"status": "ok"}')
        mock_http.request.return_value = mock_resp
        self.cls._http = mock_http
        self.cls.flush()
        ts = 1481884842
        expected = {
            'series': [
                {
                    'metric': 'prefix.runtime',
                    'points': [[ts, 123.45]],
                    'type': 'gauge',
                    'tags': ['tag1', 'tag:2']
                },
                {
                    'metric': 'prefix.svc1.limita.max_usage',
                    'points': [[ts, 0]],
                    'type': 'gauge',
                    'tags': ['tag1', 'tag:2']
                },
                {
                    'metric': 'prefix.svc1.limitb.max_usage',
                    'points': [[ts, 6]],
                    'type': 'gauge',
                    'tags': ['tag1', 'tag:2']
                },
                {
                    'metric': 'prefix.svc1.limitb.limit',
                    'points': [[ts, 10]],
                    'type': 'gauge',
                    'tags': ['tag1', 'tag:2']
                }
            ]
        }
        assert len(mock_http.mock_calls) == 1
        c = mock_http.mock_calls[0]
        assert c[0] == 'request'
        assert c[1] == (
            'POST', 'https://api.datadoghq.com/api/v1/series?api_key=myKey'
        )
        assert len(c[2]) == 2
        assert c[2]['headers'] == {'Content-type': 'application/json'}
        assert json.loads(c[2]['body'].decode()) == expected

    @freeze_time("2016-12-16 10:40:42", tz_offset=0, auto_tick_seconds=6)
    def test_api_error(self):
        self.cls._prefix = 'prefix.'
        self.cls._tags = ['tag1', 'tag:2']
        self.cls._limits = []
        self.cls._api_key = 'myKey'
        self.cls.set_run_duration(123.45)
        limA = Mock(
            name='limitA', service=Mock(service_name='SVC1')
        )
        type(limA).name = 'limitA'
        limA.get_current_usage.return_value = []
        limA.get_limit.return_value = None
        self.cls.add_limit(limA)
        limB = Mock(
            name='limitB', service=Mock(service_name='SVC1')
        )
        type(limB).name = 'limitB'
        mocku = Mock()
        mocku.get_value.return_value = 6
        limB.get_current_usage.return_value = [mocku]
        limB.get_limit.return_value = 10
        self.cls.add_limit(limB)
        mock_http = Mock()
        mock_resp = Mock(status=503, data='{"status": "NG"}')
        mock_http.request.return_value = mock_resp
        self.cls._http = mock_http
        with pytest.raises(RuntimeError) as exc:
            self.cls.flush()
        assert str(exc.value) == 'ERROR sending metrics to Datadog; API ' \
                                 'responded HTTP 503: {"status": "NG"}'
        ts = 1481884842
        expected = {
            'series': [
                {
                    'metric': 'prefix.runtime',
                    'points': [[ts, 123.45]],
                    'type': 'gauge',
                    'tags': ['tag1', 'tag:2']
                },
                {
                    'metric': 'prefix.svc1.limita.max_usage',
                    'points': [[ts, 0]],
                    'type': 'gauge',
                    'tags': ['tag1', 'tag:2']
                },
                {
                    'metric': 'prefix.svc1.limitb.max_usage',
                    'points': [[ts, 6]],
                    'type': 'gauge',
                    'tags': ['tag1', 'tag:2']
                },
                {
                    'metric': 'prefix.svc1.limitb.limit',
                    'points': [[ts, 10]],
                    'type': 'gauge',
                    'tags': ['tag1', 'tag:2']
                }
            ]
        }
        assert len(mock_http.mock_calls) == 1
        c = mock_http.mock_calls[0]
        assert c[0] == 'request'
        assert c[1] == (
            'POST', 'https://api.datadoghq.com/api/v1/series?api_key=myKey'
        )
        assert len(c[2]) == 2
        assert c[2]['headers'] == {'Content-type': 'application/json'}
        assert json.loads(c[2]['body'].decode()) == expected



"""
awslimitchecker/services/lambdafunc.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import abc  # noqa
import logging

from .base import _AwsService
from ..limit import AwsLimit
from botocore.exceptions import EndpointConnectionError

logger = logging.getLogger(__name__)


class _LambdaService(_AwsService):

    service_name = 'Lambda'
    api_name = 'lambda'

    def find_usage(self):
        """
        Determine the current usage for each limit of this service,
        and update corresponding Limit via
        :py:meth:`~.AwsLimit._add_current_usage`.
        """
        logger.debug("Getting usage for Lambda metrics")
        try:
            self.connect()
            resp = self.conn.get_account_settings()
        except EndpointConnectionError as ex:
            logger.warn('Skipping Lambda: %s', str(ex))
            return
        self.limits['Function Count']._add_current_usage(
            resp['AccountUsage']['FunctionCount']
        )
        self.limits['Total Code Size (MiB)']._add_current_usage(
            int((resp['AccountUsage']['TotalCodeSize']) / 1048576)
        )
        self._have_usage = True

    def get_limits(self):
        """
        Return all known limits for this service, as a dict of their names
        to :py:class:`~.AwsLimit` objects.

        :returns: dict of limit names to :py:class:`~.AwsLimit` objects
        :rtype: dict
        """
        logger.debug("Getting limits for Lambda")
        if self.limits != {}:
            return self.limits

        self._construct_limits()

        return self.limits

    def _update_limits_from_api(self):
        """
        Query Lambda's DescribeLimits API action, and update limits
        with the quotas returned. Updates ``self.limits``.
        """
        logger.debug("Updating limits for Lambda from the AWS API")
        if len(self.limits) == 2:
            return
        self.connect()
        lims = self.conn.get_account_settings()['AccountLimit']
        self.limits['Total Code Size (MiB)']._set_api_limit(
            (lims['TotalCodeSize'] / 1048576)
        )
        self.limits['Code Size Unzipped (MiB) per Function']._set_api_limit(
            (lims['CodeSizeUnzipped'] / 1048576)
        )
        self.limits['Unreserved Concurrent Executions']._set_api_limit(
            lims['UnreservedConcurrentExecutions']
        )
        self.limits['Concurrent Executions']._set_api_limit(
            lims['ConcurrentExecutions']
        )
        self.limits['Code Size Zipped (MiB) per Function']._set_api_limit(
            (lims['CodeSizeZipped'] / 1048576)
        )

    def _construct_limits(self):
        self.limits = {}
        self.limits['Total Code Size (MiB)'] = AwsLimit(
            'Total Code Size (MiB)',
            self,
            76800,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::Lambda::Function'
        )

        self.limits['Code Size Unzipped (MiB) per Function'] = AwsLimit(
            'Code Size Unzipped (MiB) per Function',
            self,
            250,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::Lambda::Function'
        )

        self.limits['Unreserved Concurrent Executions'] = AwsLimit(
            'Unreserved Concurrent Executions',
            self,
            1000,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::Lambda::Function'
        )

        self.limits['Concurrent Executions'] = AwsLimit(
            'Concurrent Executions',
            self,
            1000,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::Lambda::Function'
        )

        self.limits['Code Size Zipped (MiB) per Function'] = AwsLimit(
            'Code Size Zipped (MiB) per Function',
            self,
            50,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::Lambda::Function'
        )

        self.limits['Function Count'] = AwsLimit(
            'Function Count',
            self,
            None,  # unlimited
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::Lambda::Function'
        )

    def required_iam_permissions(self):
        """
        Return a list of IAM Actions required for this Service to function
        properly. All Actions will be shown with an Effect of "Allow"
        and a Resource of "*".

        :returns: list of IAM Action strings
        :rtype: list
        """
        return [
            "lambda:GetAccountSettings"
        ]



"""
awslimitchecker/tests/metrics/test_base.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2019 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from awslimitchecker.metrics.base import MetricsProvider
from awslimitchecker.metrics import Dummy, Datadog

import pytest


class MPTester(MetricsProvider):

    def flush(self):
        pass


class TestMetricsProvider(object):

    def test_init(self):
        cls = MPTester('foo')
        assert cls._region_name == 'foo'
        assert cls._duration == 0.0
        assert cls._limits == []

    def test_set_run_duration(self):
        cls = MPTester('foo')
        assert cls._duration == 0.0
        cls.set_run_duration(123.45)
        assert cls._duration == 123.45

    def test_add_limit(self):
        cls = MPTester('foo')
        assert cls._limits == []
        cls.add_limit(1)
        cls.add_limit(2)
        assert cls._limits == [1, 2]

    def test_providers_by_name(self):
        assert MetricsProvider.providers_by_name() == {
            'Dummy': Dummy,
            'MPTester': MPTester,
            'Datadog': Datadog
        }

    def test_get_provider_by_name(self):
        assert MetricsProvider.get_provider_by_name('Dummy') == Dummy

    def test_get_provider_by_name_exception(self):
        with pytest.raises(RuntimeError) as exc:
            MetricsProvider.get_provider_by_name('3993fhej')
        assert str(exc.value) == 'ERROR: "3993fhej" is not a valid ' \
                                 'MetricsProvider class name'



"""
awslimitchecker/services/cloudtrail.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import abc  # noqa
import logging

from .base import _AwsService
from ..limit import AwsLimit

logger = logging.getLogger(__name__)


class _CloudTrailService(_AwsService):

    service_name = 'CloudTrail'
    api_name = 'cloudtrail'
    aws_type = 'AWS::CloudTrail::Trail'

    def find_usage(self):
        """
        Determine the current usage for each limit of this service,
        and update corresponding Limit via
        :py:meth:`~.AwsLimit._add_current_usage`.
        """
        logger.debug("Checking usage for service %s", self.service_name)

        self.connect()
        for lim in self.limits.values():
            lim._reset_usage()

        self._find_usage_cloudtrail()
        self._have_usage = True
        logger.debug("Done checking usage.")

    def _find_usage_cloudtrail(self):
        """Calculate current usage for CloudTrail related metrics"""

        trail_list = self.conn.describe_trails(
            includeShadowTrails=False
        )['trailList']
        trail_count = len(trail_list) if trail_list else 0

        for trail in trail_list:
            data_resource_count = 0
            if self.conn._client_config.region_name == trail['HomeRegion']:
                try:
                    response = self.conn.get_event_selectors(
                        TrailName=trail['TrailARN']
                    )
                except Exception as ex:
                    logger.debug(
                        'Unable to call GetEventSelectors on CloudTrail trail '
                        '%s: %s', trail, ex
                    )
                    continue
                event_selectors = response['EventSelectors']
                for event_selector in event_selectors:
                    data_resource_count += len(
                        event_selector.get('DataResources', [])
                    )
                self.limits['Event Selectors Per Trail']._add_current_usage(
                    len(event_selectors),
                    aws_type='AWS::CloudTrail::EventSelector',
                    resource_id=trail['Name']
                )
                self.limits['Data Resources Per Trail']._add_current_usage(
                    data_resource_count,
                    aws_type='AWS::CloudTrail::DataResource',
                    resource_id=trail['Name']
                )
            else:
                logger.debug(
                    'Ignoring event selectors and data resources for '
                    'CloudTrail %s in non-home region' % trail['Name']
                )
        self.limits['Trails Per Region']._add_current_usage(
            trail_count,
            aws_type=self.aws_type
        )

    def get_limits(self):
        """
        Return all known limits for this service, as a dict of their names
        to :py:class:`~.AwsLimit` objects.

        :returns: dict of limit names to :py:class:`~.AwsLimit` objects
        :rtype: dict
        """
        logger.debug("Gathering %s's limits from AWS", self.service_name)

        if self.limits:
            return self.limits
        limits = {}

        limits['Trails Per Region'] = AwsLimit(
            'Trails Per Region',
            self,
            5,
            self.warning_threshold,
            self.critical_threshold,
            limit_type=self.aws_type
        )

        limits['Event Selectors Per Trail'] = AwsLimit(
            'Event Selectors Per Trail',
            self,
            5,
            self.warning_threshold,
            self.critical_threshold,
            limit_type=self.aws_type,
            limit_subtype='AWS::CloudTrail::EventSelector'
        )

        limits['Data Resources Per Trail'] = AwsLimit(
            'Data Resources Per Trail',
            self,
            250,
            self.warning_threshold,
            self.critical_threshold,
            limit_type=self.aws_type,
            limit_subtype='AWS::CloudTrail::DataResource'
        )

        self.limits = limits
        return limits

    def required_iam_permissions(self):
        """
        Return a list of IAM Actions required for this Service to function
        properly. All Actions will be shown with an Effect of "Allow"
        and a Resource of "*".

        :returns: list of IAM Action strings
        :rtype: list
        """
        return [
            "cloudtrail:DescribeTrails",
            "cloudtrail:GetEventSelectors",
        ]



"""
awslimitchecker/tests/services/test_apigateway.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
from copy import deepcopy
from awslimitchecker.tests.services import result_fixtures
from awslimitchecker.services.apigateway import _ApigatewayService

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import patch, call, Mock, DEFAULT
else:
    from unittest.mock import patch, call, Mock, DEFAULT


pbm = 'awslimitchecker.services.apigateway'  # module patch base
pb = '%s._ApigatewayService' % pbm  # class patch pase


class Test_ApigatewayService(object):

    def test_init(self):
        """test __init__()"""
        cls = _ApigatewayService(21, 43, {}, None)
        assert cls.service_name == 'ApiGateway'
        assert cls.api_name == 'apigateway'
        assert cls.conn is None
        assert cls.warning_threshold == 21
        assert cls.critical_threshold == 43
        assert cls.quotas_service_code == 'apigateway'

    def test_get_limits(self):
        cls = _ApigatewayService(21, 43, {}, None)
        cls.limits = {}
        res = cls.get_limits()
        assert sorted(res.keys()) == sorted([
            'Regional APIs per account',
            'Private APIs per account',
            'Edge APIs per account',
            'API keys per account',
            'Client certificates per account',
            'Custom authorizers per API',
            'Documentation parts per API',
            'Resources per API',
            'Stages per API',
            'Usage plans per account',
            'VPC Links per account'
        ])
        for name, limit in res.items():
            assert limit.service == cls
            assert limit.def_warning_threshold == 21
            assert limit.def_critical_threshold == 43
        assert res['VPC Links per account'].quota_name == 'VPC links'

    def test_get_limits_again(self):
        """test that existing limits dict is returned on subsequent calls"""
        mock_limits = Mock()
        cls = _ApigatewayService(21, 43, {}, None)
        cls.limits = mock_limits
        res = cls.get_limits()
        assert res == mock_limits

    def test_find_usage(self):
        mock_conn = Mock()
        with patch('%s.connect' % pb) as mock_connect:
            with patch.multiple(
                pb,
                autospec=True,
                _find_usage_apis=DEFAULT,
                _find_usage_api_keys=DEFAULT,
                _find_usage_certs=DEFAULT,
                _find_usage_plans=DEFAULT,
                _find_usage_vpc_links=DEFAULT
            ) as mocks:
                cls = _ApigatewayService(21, 43, {}, None)
                cls.conn = mock_conn
                assert cls._have_usage is False
                cls.find_usage()
        assert mock_connect.mock_calls == [call()]
        assert cls._have_usage is True
        assert mock_conn.mock_calls == []
        assert mocks['_find_usage_apis'].mock_calls == [call(cls)]
        assert mocks['_find_usage_api_keys'].mock_calls == [call(cls)]
        assert mocks['_find_usage_certs'].mock_calls == [call(cls)]
        assert mocks['_find_usage_plans'].mock_calls == [call(cls)]
        assert mocks['_find_usage_vpc_links'].mock_calls == [call(cls)]

    def test_find_usage_apis(self):
        mock_conn = Mock()
        res = result_fixtures.ApiGateway.get_rest_apis
        mock_paginator = Mock()
        mock_paginator.paginate.return_value = res

        def se_res_paginate(restApiId=None):
            return result_fixtures.ApiGateway.get_resources[restApiId]

        mock_res_paginator = Mock()
        mock_res_paginator.paginate.side_effect = se_res_paginate

        def se_get_paginator(api_name):
            if api_name == 'get_rest_apis':
                return mock_paginator
            elif api_name == 'get_resources':
                return mock_res_paginator

        def se_paginate_dict(*args, **kwargs):
            if args[0] == mock_conn.get_documentation_parts:
                return result_fixtures.ApiGateway.doc_parts[kwargs['restApiId']]
            if args[0] == mock_conn.get_authorizers:
                return result_fixtures.ApiGateway.authorizers[
                    kwargs['restApiId']
                ]

        def se_get_stages(restApiId=None):
            return result_fixtures.ApiGateway.stages[restApiId]

        mock_conn.get_paginator.side_effect = se_get_paginator
        mock_conn.get_stages.side_effect = se_get_stages
        cls = _ApigatewayService(21, 43, {}, None)
        cls.conn = mock_conn
        with patch('%s.paginate_dict' % pbm, autospec=True) as mock_pd:
            with patch('%s.logger' % pbm) as mock_logger:
                mock_pd.side_effect = se_paginate_dict
                cls._find_usage_apis()
        # APIs usage
        usage = cls.limits['Regional APIs per account'].get_current_usage()
        assert len(usage) == 1
        assert usage[0].get_value() == 2
        usage = cls.limits['Edge APIs per account'].get_current_usage()
        assert len(usage) == 1
        assert usage[0].get_value() == 2
        usage = cls.limits['Private APIs per account'].get_current_usage()
        assert len(usage) == 1
        assert usage[0].get_value() == 1
        # Resources usage
        usage = cls.limits['Resources per API'].get_current_usage()
        assert len(usage) == 5
        assert usage[0].resource_id == 'api3'
        assert usage[0].get_value() == 0
        assert usage[1].resource_id == 'api2'
        assert usage[1].get_value() == 2
        assert usage[2].resource_id == 'api1'
        assert usage[2].get_value() == 3
        assert usage[3].resource_id == 'api4'
        assert usage[3].get_value() == 0
        assert usage[4].resource_id == 'api5'
        assert usage[4].get_value() == 0
        usage = cls.limits['Documentation parts per API'].get_current_usage()
        assert len(usage) == 5
        assert usage[0].resource_id == 'api3'
        assert usage[0].get_value() == 2
        assert usage[1].resource_id == 'api2'
        assert usage[1].get_value() == 1
        assert usage[2].resource_id == 'api1'
        assert usage[2].get_value() == 4
        assert usage[3].resource_id == 'api4'
        assert usage[3].get_value() == 1
        assert usage[4].resource_id == 'api5'
        assert usage[4].get_value() == 1
        usage = cls.limits['Stages per API'].get_current_usage()
        assert len(usage) == 5
        assert usage[0].resource_id == 'api3'
        assert usage[0].get_value() == 2
        assert usage[1].resource_id == 'api2'
        assert usage[1].get_value() == 1
        assert usage[2].resource_id == 'api1'
        assert usage[2].get_value() == 3
        assert usage[3].resource_id == 'api4'
        assert usage[3].get_value() == 1
        assert usage[4].resource_id == 'api5'
        assert usage[4].get_value() == 1
        usage = cls.limits['Custom authorizers per API'].get_current_usage()
        assert len(usage) == 5
        assert usage[0].resource_id == 'api3'
        assert usage[0].get_value() == 0
        assert usage[1].resource_id == 'api2'
        assert usage[1].get_value() == 2
        assert usage[2].resource_id == 'api1'
        assert usage[2].get_value() == 1
        assert usage[3].resource_id == 'api4'
        assert usage[3].get_value() == 0
        assert usage[4].resource_id == 'api5'
        assert usage[4].get_value() == 0
        assert mock_conn.mock_calls == [
            call.get_paginator('get_rest_apis'),
            call.get_paginator('get_resources'),
            call.get_stages(restApiId='api3'),
            call.get_paginator('get_resources'),
            call.get_stages(restApiId='api2'),
            call.get_paginator('get_resources'),
            call.get_stages(restApiId='api1'),
            call.get_paginator('get_resources'),
            call.get_stages(restApiId='api4'),
            call.get_paginator('get_resources'),
            call.get_stages(restApiId='api5')
        ]
        assert mock_paginator.mock_calls == [call.paginate()]
        assert mock_res_paginator.mock_calls == [
            call.paginate(restApiId='api3'),
            call.paginate(restApiId='api2'),
            call.paginate(restApiId='api1'),
            call.paginate(restApiId='api4'),
            call.paginate(restApiId='api5')
        ]
        assert mock_pd.mock_calls == [
            call(
                mock_conn.get_documentation_parts,
                restApiId='api3',
                alc_marker_path=['position'],
                alc_data_path=['items'],
                alc_marker_param='position'
            ),
            call(
                mock_conn.get_authorizers,
                restApiId='api3',
                alc_marker_path=['position'],
                alc_data_path=['items'],
                alc_marker_param='position'
            ),
            call(
                mock_conn.get_documentation_parts,
                restApiId='api2',
                alc_marker_path=['position'],
                alc_data_path=['items'],
                alc_marker_param='position'
            ),
            call(
                mock_conn.get_authorizers,
                restApiId='api2',
                alc_marker_path=['position'],
                alc_data_path=['items'],
                alc_marker_param='position'
            ),
            call(
                mock_conn.get_documentation_parts,
                restApiId='api1',
                alc_marker_path=['position'],
                alc_data_path=['items'],
                alc_marker_param='position'
            ),
            call(
                mock_conn.get_authorizers,
                restApiId='api1',
                alc_marker_path=['position'],
                alc_data_path=['items'],
                alc_marker_param='position'
            ),
            call(
                mock_conn.get_documentation_parts,
                restApiId='api4',
                alc_marker_path=['position'],
                alc_data_path=['items'],
                alc_marker_param='position'
            ),
            call(
                mock_conn.get_authorizers,
                restApiId='api4',
                alc_marker_path=['position'],
                alc_data_path=['items'],
                alc_marker_param='position'
            ),
            call(
                mock_conn.get_documentation_parts,
                restApiId='api5',
                alc_marker_path=['position'],
                alc_data_path=['items'],
                alc_marker_param='position'
            ),
            call(
                mock_conn.get_authorizers,
                restApiId='api5',
                alc_marker_path=['position'],
                alc_data_path=['items'],
                alc_marker_param='position'
            )
        ]
        assert mock_logger.mock_calls == [
            call.debug('Finding usage for APIs'),
            call.debug('Found %d APIs', 5),
            call.debug('Finding usage for per-API limits')
        ]

    def test_find_usage_apis_stages_now_paginated(self):
        mock_conn = Mock()
        res = result_fixtures.ApiGateway.get_rest_apis
        mock_paginator = Mock()
        mock_paginator.paginate.return_value = res

        def se_res_paginate(restApiId=None):
            return result_fixtures.ApiGateway.get_resources[restApiId]

        mock_res_paginator = Mock()
        mock_res_paginator.paginate.side_effect = se_res_paginate

        def se_get_paginator(api_name):
            if api_name == 'get_rest_apis':
                return mock_paginator
            elif api_name == 'get_resources':
                return mock_res_paginator

        def se_paginate_dict(*args, **kwargs):
            if args[0] == mock_conn.get_documentation_parts:
                return result_fixtures.ApiGateway.doc_parts[kwargs['restApiId']]
            if args[0] == mock_conn.get_authorizers:
                return result_fixtures.ApiGateway.authorizers[
                    kwargs['restApiId']
                ]

        def se_get_stages(restApiId=None):
            r = deepcopy(result_fixtures.ApiGateway.stages[restApiId])
            r['position'] = 'foo'
            return r

        mock_conn.get_paginator.side_effect = se_get_paginator
        mock_conn.get_stages.side_effect = se_get_stages
        cls = _ApigatewayService(21, 43, {}, None)
        cls.conn = mock_conn
        with patch('%s.paginate_dict' % pbm, autospec=True) as mock_pd:
            with patch('%s.logger' % pbm) as mock_logger:
                mock_pd.side_effect = se_paginate_dict
                cls._find_usage_apis()
        assert mock_logger.mock_calls == [
            call.debug('Finding usage for APIs'),
            call.debug('Found %d APIs', 5),
            call.debug('Finding usage for per-API limits'),
            call.warning(
                'APIGateway get_stages returned more keys than present in '
                'boto3 docs: %s', ['item', 'position']
            )
        ]

    def test_find_usage_plans(self):
        mock_conn = Mock()
        res = result_fixtures.ApiGateway.plans
        mock_paginator = Mock()
        mock_paginator.paginate.return_value = res

        mock_conn.get_paginator.return_value = mock_paginator
        cls = _ApigatewayService(21, 43, {}, None)
        cls.conn = mock_conn
        with patch('%s.logger' % pbm) as mock_logger:
            cls._find_usage_plans()
        # APIs usage
        usage = cls.limits['Usage plans per account'].get_current_usage()
        assert len(usage) == 1
        assert usage[0].get_value() == 4
        assert mock_conn.mock_calls == [
            call.get_paginator('get_usage_plans'),
            call.get_paginator().paginate()
        ]
        assert mock_paginator.mock_calls == [call.paginate()]
        assert mock_logger.mock_calls == [
            call.debug('Finding usage for Usage Plans')
        ]

    def test_find_usage_certs(self):
        mock_conn = Mock()
        res = result_fixtures.ApiGateway.certs
        mock_paginator = Mock()
        mock_paginator.paginate.return_value = res

        mock_conn.get_paginator.return_value = mock_paginator
        cls = _ApigatewayService(21, 43, {}, None)
        cls.conn = mock_conn
        with patch('%s.logger' % pbm) as mock_logger:
            cls._find_usage_certs()
        # APIs usage
        usage = cls.limits[
            'Client certificates per account'].get_current_usage()
        assert len(usage) == 1
        assert usage[0].get_value() == 2
        assert mock_conn.mock_calls == [
            call.get_paginator('get_client_certificates'),
            call.get_paginator().paginate()
        ]
        assert mock_paginator.mock_calls == [call.paginate()]
        assert mock_logger.mock_calls == [
            call.debug('Finding usage for Client Certificates')
        ]

    def test_find_usage_api_keys(self):
        mock_conn = Mock()
        res = result_fixtures.ApiGateway.api_keys
        mock_paginator = Mock()
        mock_paginator.paginate.return_value = res

        mock_conn.get_paginator.return_value = mock_paginator
        cls = _ApigatewayService(21, 43, {}, None)
        cls.conn = mock_conn
        with patch('%s.logger' % pbm) as mock_logger:
            cls._find_usage_api_keys()
        # API Keys usage
        usage = cls.limits[
            'API keys per account'].get_current_usage()
        assert len(usage) == 1
        assert usage[0].get_value() == 4
        assert mock_conn.mock_calls == [
            call.get_paginator('get_api_keys'),
            call.get_paginator().paginate()
        ]
        assert mock_paginator.mock_calls == [call.paginate()]
        assert mock_logger.mock_calls == [
            call.debug('Finding usage for API Keys')
        ]

    def test_find_usage_vpc_links(self):
        mock_conn = Mock()
        res = result_fixtures.ApiGateway.vpc_links
        mock_paginator = Mock()
        mock_paginator.paginate.return_value = res

        mock_conn.get_paginator.return_value = mock_paginator
        cls = _ApigatewayService(21, 43, {}, None)
        cls.conn = mock_conn
        with patch('%s.logger' % pbm) as mock_logger:
            cls._find_usage_vpc_links()
        # APIs usage
        usage = cls.limits[
            'VPC Links per account'].get_current_usage()
        assert len(usage) == 1
        assert usage[0].get_value() == 3
        assert mock_conn.mock_calls == [
            call.get_paginator('get_vpc_links'),
            call.get_paginator().paginate()
        ]
        assert mock_paginator.mock_calls == [call.paginate()]
        assert mock_logger.mock_calls == [
            call.debug('Finding usage for VPC Links')
        ]

    def test_required_iam_permissions(self):
        cls = _ApigatewayService(21, 43, {}, None)
        assert cls.required_iam_permissions() == [
            "apigateway:GET",
            "apigateway:HEAD",
            "apigateway:OPTIONS"
        ]
